__all__ = ('cr_pg_channel_object', 'create_partial_channel_from_data', 'create_partial_channel_from_id')

import warnings
from functools import partial as partial_func

from scarletio import export, include

from ...core import CHANNELS

from ..channel_metadata.base import CHANNEL_METADATA_ICON_SLOT
from ..channel_metadata.fields import (
    put_applied_tag_ids_into, put_auto_archive_after_into, put_bitrate_into, put_default_forum_layout_into,
    put_default_sort_order_into, put_default_thread_auto_archive_after_into, put_default_thread_reaction_into,
    put_default_thread_slowmode_into, put_flags_into, put_invitable_into, put_name_into, put_nsfw_into, put_open_into,
    put_parent_id_into, put_permission_overwrites_into, put_position_into, put_region_into, put_slowmode_into,
    put_topic_into, put_user_limit_into, put_video_quality_mode_into, validate_applied_tag_ids,
    validate_auto_archive_after, validate_bitrate, validate_default_forum_layout, validate_default_sort_order,
    validate_default_thread_auto_archive_after, validate_default_thread_reaction, validate_default_thread_slowmode,
    validate_flags, validate_invitable, validate_name, validate_nsfw, validate_open, validate_parent_id,
    validate_permission_overwrites, validate_position, validate_region, validate_slowmode, validate_topic,
    validate_user_limit, validate_video_quality_mode
)

from .preinstanced import ChannelType


Channel = include('Channel')


CHANNEL_GUILD_MAIN_FIELD_CONVERTERS = {
    'bitrate': (validate_bitrate, put_bitrate_into),
    'default_forum_layout': (validate_default_forum_layout, put_default_forum_layout_into),
    'default_sort_order': (validate_default_sort_order, put_default_sort_order_into),
    'default_thread_auto_archive_after': (
        validate_default_thread_auto_archive_after, put_default_thread_auto_archive_after_into
    ),
    'default_thread_reaction': (validate_default_thread_reaction, put_default_thread_reaction_into),
    'default_thread_slowmode': (validate_default_thread_slowmode, put_default_thread_slowmode_into),
    'flags': (validate_flags, put_flags_into),
    'name': (validate_name, put_name_into),
    'nsfw': (validate_nsfw, put_nsfw_into),
    'parent_id': (validate_parent_id, put_parent_id_into),
    'permission_overwrites': (validate_permission_overwrites, put_permission_overwrites_into),
    'position': (validate_position, put_position_into),
    'region': (validate_region, put_region_into),
    'slowmode': (validate_slowmode, put_slowmode_into),
    'topic': (validate_topic, put_topic_into),
    'user_limit': (validate_user_limit, put_user_limit_into),
    'video_quality_mode': (validate_video_quality_mode, put_video_quality_mode_into),
}


CHANNEL_PRIVATE_GROUP_FIELD_CONVERTERS = {
    'icon': (
        CHANNEL_METADATA_ICON_SLOT.validate_icon,
        partial_func(CHANNEL_METADATA_ICON_SLOT.put_into, as_data = True),
    ),
    'name': (validate_name, put_name_into),
}


CHANNEL_GUILD_THREAD_FIELD_CONVERTERS = {
    'applied_tag_ids': (validate_applied_tag_ids, put_applied_tag_ids_into),
    'auto_archive_after': (
        validate_auto_archive_after, partial_func(put_auto_archive_after_into, flatten_thread_metadata = True)
    ),
    'flags': (validate_flags, put_flags_into),
    'invitable': (validate_invitable, partial_func(put_invitable_into, flatten_thread_metadata = True)),
    'name': (validate_name, put_name_into),
    'open_': (validate_open, partial_func(put_open_into, flatten_thread_metadata = True)),
    'slowmode': (validate_slowmode, put_slowmode_into),
}


CHANNEL_GUILD_FIELD_CONVERTERS = {
    **CHANNEL_GUILD_THREAD_FIELD_CONVERTERS,
    **CHANNEL_GUILD_MAIN_FIELD_CONVERTERS,
}


def create_partial_channel_from_data(data, guild_id):
    """
    Creates a partial channel from partial channel data.
    
    Parameters
    ----------
    data : `None`, `dict` of (`str`, `Any`) items
        Partial channel data received from Discord.
    guild_id : `int`
        The channel's guild's identifier.
    
    Returns
    -------
    channel : `None`, ``Channel``
        The created partial channel, or `None`, if no data was received.
    """
    if (data is None) or (not data):
        return None
    
    channel_id = int(data['id'])
    try:
        return CHANNELS[channel_id]
    except KeyError:
        pass
    
    channel = Channel._from_partial_data(data, channel_id, guild_id)
    CHANNELS[channel_id] = channel
    
    return channel


@export
def create_partial_channel_from_id(channel_id, channel_type, guild_id):
    """
    Creates a new partial channel from the given identifier.
    
    Parameters
    ----------
    channel_id : `int`
        The channel's identifier.
    channel_type : ``ChannelType``
        The channel's type.
    guild_id : `int`
        A guild's identifier of the created channel.
    """
    try:
        return CHANNELS[channel_id]
    except KeyError:
        pass
    
    channel = Channel._create_empty(channel_id, channel_type, guild_id)
    CHANNELS[channel_id] = channel
    
    return channel


HAS_SLOWMODE = (
    *(
        channel_type for channel_type in ChannelType.INSTANCES.values()
        if channel_type.flags.guild and channel_type.flags.textual
    ),
    ChannelType.guild_forum,
)



def cr_pg_channel_object(name, channel_type, *, guild = None, **keyword_parameters):
    """
    Deprecated, please use `Channel(..).to_data(...)` instead.
    
    Will be removed in 2023 February.
    """
    warnings.warn(
        (
            f'`cr_pg_channel_object` is deprecated and will be removed in 2023 February. '
            f'Please use `Channel(..).to_data(...)` instead.'
        ),
        FutureWarning,
        stacklevel = 2,
    )
    return Channel(channel_type = channel_type, name = name, **keyword_parameters).to_data()



__all__ = ()

from .application_command import APPLICATION_COMMAND_CONVERTERS
from .auto_moderation_rule import AUTO_MODERATION_RULE_CONVERTERS
from .channel import CHANNEL_CONVERTERS
from .channel_permission_overwrite import CHANNEL_PERMISSION_OVERWRITE_CONVERTERS
from .emoji import EMOJI_CONVERTERS
from .guild import GUILD_CONVERTERS
from .integration import INTEGRATION_CONVERTERS
from .invite import INVITE_CONVERTERS
from .role import ROLE_CONVERTERS
from .scheduled_event import SCHEDULED_EVENT_CONVERTERS
from .stage import STAGE_CONVERTERS
from .sticker import STICKER_CONVERTERS
from .user import USER_CONVERTERS
from .webhook import WEBHOOK_CONVERTERS


def merge_converters(*converters_to_merge):
    """
    Merges the given converter groups.
    
    Parameters
    ----------
    converters_to_merge : `dict` of (`str`, `FunctionType`) items
        Audit log change converters.
    
    Returns
    -------
    merged_converters : `dict` of (`str`, `FunctionType`) items
    """
    multi_type_converters = set()
    merged_converters = {}
    
    for converters in converters_to_merge:
        for key, converter in converters.items():
            if key in multi_type_converters:
                continue
            
            try:
                actual_converter = merged_converters[key]
            except KeyError:
                merged_converters[key] = converter
                continue
            
            if actual_converter is converter:
                continue
            
            del merged_converters[key]
            multi_type_converters.add(key)
            continue
    
    return merged_converters


MERGED_CONVERTERS = merge_converters(
    APPLICATION_COMMAND_CONVERTERS,
    CHANNEL_CONVERTERS,
    AUTO_MODERATION_RULE_CONVERTERS,
    CHANNEL_PERMISSION_OVERWRITE_CONVERTERS,
    EMOJI_CONVERTERS,
    GUILD_CONVERTERS,
    INTEGRATION_CONVERTERS,
    INVITE_CONVERTERS,
    ROLE_CONVERTERS,
    SCHEDULED_EVENT_CONVERTERS,
    STAGE_CONVERTERS,
    STICKER_CONVERTERS,
    USER_CONVERTERS,
    WEBHOOK_CONVERTERS,
)



__all__ = ()

import warnings

from scarletio import Compound

from ...channel import Channel, ChannelType, ForumTag, PermissionOverwrite, create_partial_channel_from_id
from ...channel.channel.utils import (
    CHANNEL_GUILD_FIELD_CONVERTERS, CHANNEL_GUILD_MAIN_FIELD_CONVERTERS, CHANNEL_PRIVATE_GROUP_FIELD_CONVERTERS
)
from ...channel.permission_overwrite.utils import (
    PERMISSION_OVERWRITE_FIELD_CONVERTERS, PERMISSION_OVERWRITE_PERMISSION_FIELD_CONVERTERS
)
from ...channel.forum_tag.utils import FORUM_TAG_FIELD_CONVERTERS
from ...guild import Guild, create_partial_guild_from_id
from ...http import DiscordHTTPClient
from ...payload_building import build_create_payload, build_edit_payload
from ...webhook import Webhook

from ..functionality_helpers import channel_move_sort_key
from ..request_helpers import (
    get_channel_and_id, get_channel_id, get_forum_tag_and_id, get_forum_tag_id, get_guild_and_id, get_guild_id,
    get_permission_overwrite_target_id, get_user_id
)

def _assert__channel_group_create__users(users):
    """
    Asserts the `users` parameter of ``Client.channel_group_create`.
    
    Parameters
    ----------
    users : `tuple` of ``ClientUserBase``, `int`
        The users to create the channel with.
    
    Raises
    ------
    AssertionError
        - If the total amount of users is less than `2`.
    """
    if len(users) < 2:
        raise AssertionError(
            f'group channel can be created at least with at least `2` users,  got '
            f'{len(users)}; {users!r}.'
        )
    
    return True


def _forum_tag_data_array_sort_key(forum_tag_data):
    """
    Sort key used to sort forum tag data.
    
    Parameters
    ----------
    forum_tag_data : `dict` of (`str`, `Any`) items
        Forum tag data.
    
    Returns
    -------
    sort_key : `int`
    """
    return int(forum_tag_data['id'])


class ClientCompoundChannelEndpoints(Compound):
    
    http : DiscordHTTPClient
    id : int
    bot : bool
    private_channels : dict
    
    
    async def channel_group_leave(self, channel):
        """
        Leaves the client from the specified group channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to leave from.
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_private_group)
        
        await self.http.channel_group_leave(channel_id)
    
    
    async def channel_group_user_add(self, channel, *users):
        """
        Adds the users to the given group channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to add the `users` to.
        *users : ``ClientUserBase``, `int`
            The users to add to the `channel`.
        
        Raises
        ------
        TypeError
            - If `channel` was not given neither as ``Channel`` nor `int`.
            - If `users` contains non ``ClientUserBase``, neither `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_private_group)
        
        user_ids = set()
        for user in users:
            user_id = get_user_id(user)
            user_ids.add(user_id)
        
        for user_id in user_ids:
            await self.http.channel_group_user_add(channel_id, user_id)
    
    
    async def channel_group_user_delete(self, channel, *users):
        """
        Removes the users from the given group channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel from where the `users` will be removed.
        *users : ``ClientUserBase``, `int`
            The users to remove from the `channel`.
        
        Raises
        ------
        TypeError
            - If `channel` was not given neither as ``Channel`` nor `int`.
            - If `users` contains non ``ClientUserBase``, neither `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_private_group)
        
        user_ids = set()
        for user in users:
            user_id = get_user_id(user)
            user_ids.add(user_id)
        
        for user_id in user_ids:
            await self.http.channel_group_user_delete(channel_id, user_id)
    
    
    async def channel_group_edit(self, channel, channel_template = None, **keyword_parameters):
        """
        Edits the given group channel. Only the provided parameters will be edited.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to edit.
        
        channel_template : `None`, ``Channel`` = `None`, Optional
            A channel to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to define which fields should be modified.
        
        Other Parameters
        ----------------
        icon : `None`, `bytes-like`, Optional (Keyword only)
            The new icon of the channel. By passing `None` your can remove the actual one.
        
        name : `None`, `str`, Optional (Keyword only)
            The new name of the channel. By passing `None` or an empty string you can remove the actual one.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        Notes
        -----
        No request is done if no optional parameter is provided.
        """
        channel_id = get_channel_id(channel, Channel.is_private_group)
        data = build_edit_payload(channel, channel_template, CHANNEL_PRIVATE_GROUP_FIELD_CONVERTERS, keyword_parameters)

        if data:
            await self.http.channel_group_edit(channel_id, data)
    
    
    async def channel_group_create(self, *users):
        """
        Creates a group channel with the given users.
        
        This method is a coroutine.
        
        Parameters
        ----------
        *users : ``ClientUserBase``, `int`
            The users to create the channel with.
        
        Returns
        -------
        channel : ``Channel``
            The created group channel.
        
        Raises
        ------
        TypeError
            If `users` contain not only ``User`, ``Client``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        This endpoint does not support bot accounts.
        """
        assert _assert__channel_group_create__users(users)
        
        user_ids = set()
        for user in users:
            user_id = get_user_id(user)
            user_ids.add(user_id)
        
        user_ids.add(self.id)
        
        data = {'recipients': user_ids}
        data = await self.http.channel_group_create(self.id, data)
        return Channel.from_data(data, self, 0)
    
    
    async def channel_private_create(self, user):
        """
        Creates a private channel with the given user. If there is an already cached private channel with the user,
        returns that.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ``ClientUserBase``, `int`
            The user to create the private with.
        
        Returns
        -------
        channel : ``Channel``
            The created private channel.
        
        Raises
        ------
        TypeError
            If `user` was not given neither as ``ClientUserBase`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        user_id = get_user_id(user)
        
        try:
            channel = self.private_channels[user_id]
        except KeyError:
            data = await self.http.channel_private_create({'recipient_id': user_id})
            channel = Channel.from_data(data, self, 0)
        
        return channel
    
    
    async def channel_private_get_all(self):
        """
        Request the client's private + group channels and returns them in a list. At the case of bot accounts the
        request returns an empty list, so we skip it.
        
        This method is a coroutine.
        
        Returns
        -------
        channels : `list` of (``Channel``, ``Channel``) objects
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channels = []
        if (not self.bot):
            data = await self.http.channel_private_get_all()
            for channel_data in data:
                channel = Channel.from_data(channel_data, self, 0)
                channels.append(channel)
        
        return channels
    
    
    async def channel_move(self, channel, visual_position, *, parent=..., lock_permissions = False, reason = None):
        """
        Moves a guild channel to the given visual position under it's parent, or guild. If the algorithm can not
        place the channel exactly on that location, it will place it as close, as it can. If there is nothing to
        move, then the request is skipped.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``
            The channel to be moved.
        visual_position : `int`
            The visual position where the channel should go.
        parent : `None`, ``Channel``, Optional (Keyword only)
            If not set, then the channel will keep it's current parent. If the parameter is set ``Guild`` or to
            `None`, then the  channel will be moved under the guild itself, Or if passed as ``Channel``,
            then the channel will be moved under it.
        lock_permissions : `bool` = `False`, Optional (Keyword only)
            If you want to sync the permissions with the new category set it to `True`. Defaults to `False`.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        ValueError
            - If the `channel` would be between guilds.
            - If parent channel would be moved under an other category.
        TypeError
            - If `channel` was isn ot movable.
            - If `parent` was not given as `None`, ``Channel``.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        This method also fixes the messy channel positions of Discord to an intuitive one.
        """
        # Check channel type
        if (not channel.is_in_group_guild_sortable()) and (not channel.partial):
            raise TypeError(
                f'`channel` can be any movable guild channel, got {channel.__class__.__name__}; {channel!r}.'
            )
        
        # Check whether the channel is partial.
        guild = channel.guild
        if guild is None:
            # Cannot move partial channels, leave
            return
        
        # Check parent
        if parent is ...:
            parent = channel.parent
        elif parent is None:
            parent = None
        elif isinstance(parent, Channel):
            if parent.guild is not guild:
                raise ValueError(
                    f'Can not move channel between guilds! Channel\'s guild: {guild!r}; Category\'s '
                    f'guild: {parent.guild!r}'
                )
        else:
            raise TypeError(
                f'`parent` can be `None` or {Channel.__name__}`, got {parent.__class__.__name__}; {parent!r}.'
            )
        
        # Cannot put category under category
        if isinstance(parent, Channel) and channel.is_guild_category() and (not channel.partial):
            raise ValueError(
                f'Can not move category channel under category channel. channel = {channel!r}; parent = {parent!r}'
            )
        
        if not isinstance(visual_position, int):
            raise TypeError(
                f'`visual_position` can be `int`, got {visual_position.__class__.__name__}; '
                f'{visual_position!r}.'
            )
        
        if not isinstance(lock_permissions, bool):
            raise TypeError(
                f'`lock_permissions` can be `bool`, got {lock_permissions.__class__.__name__}; {lock_permissions!r}.'
            )
        
        # Cap at 0
        if visual_position < 0:
            visual_position = 0
        
        # If the channel is where it should be, we can leave.
        if (parent is not None):
            # Add `parent is not None` check for the linter
            if (channel.parent is parent) and (parent.channel_list.index(channel) == visual_position):
                return
        
        # Create a display state, where each channel is listed.
        # Categories are inside of a tuple, where they are the first element of it and their channels are the second.
        display_state = guild.channel_list
        
        for index in range(len(display_state)):
            iter_channel = display_state[index]
            if iter_channel.is_guild_category():
                display_state[index] = iter_channel, iter_channel.channel_list
        
        # Generate a state where the channels are theoretically ordered with tuples
        display_new = []
        for iter_channel in display_state:
            if isinstance(iter_channel, tuple):
                iter_channel, sub_channels = iter_channel
                display_sub_channels = []
                for sub_channel in sub_channels:
                    channel_key = (sub_channel.order_group, sub_channel.position, sub_channel.id, None)
                    display_sub_channels.append(channel_key)
            else:
                display_sub_channels = None
            
            channel_key = (iter_channel.order_group, iter_channel.position, iter_channel.id, display_sub_channels)
            
            display_new.append(channel_key)
        
        # We have 2 display states, we will compare the old to the new one when calculating differences, but we didn't
        # move our channel yet!
        
        # We get from where we will move from.
        old_parent = channel.parent
        if isinstance(old_parent, Guild):
            move_from = display_new
        else:
            old_parent_id = old_parent.id
            for channel_key in display_new:
                if channel_key[2] == old_parent_id:
                    move_from = channel_key[3]
                    break
            
            else:
                # If no breaking was not done, our channel not exists, lol
                return
        
        # We got from which thing we will move from, so we remove first
        
        channel_id = channel.id
        
        for index in range(len(move_from)):
            channel_key = move_from[index]
            
            if channel_key[2] == channel_id:
                channel_key_to_move = channel_key
                del move_from[index]
                break
        
        else:
            # If breaking was not done, our channel not exists, lol
            return
        
        # We get to where we will move to.
        if parent is None:
            move_to = display_new
        else:
            new_parent_id = parent.id
            for channel_key in display_new:
                if channel_key[2] == new_parent_id:
                    move_to = channel_key[3]
                    break
            
            else:
                # If no breaking was not done, our channel not exists, lol
                return
        
        # Move, yayyy
        move_to.insert(visual_position, channel_key_to_move)
        # Reorder
        move_to.sort(key = channel_move_sort_key)
        
        # Now we resort every channel in the guild and categories, mostly for security issues
        to_sort_all = [display_new]
        for channel_key in display_new:
            display_sub_channels = channel_key[3]
            if display_sub_channels is not None:
                to_sort_all.append(display_sub_channels)
        
        ordered = []
        
        for to_sort in to_sort_all:
            expected_channel_order_group = 0
            channel_position = 0
            for sort_key in to_sort:
                channel_order_group = sort_key[0]
                channel_id = sort_key[2]
                
                if channel_order_group != expected_channel_order_group:
                    expected_channel_order_group = channel_order_group
                    channel_position = 0
                
                ordered.append((channel_position, channel_id))
                channel_position += 1
                continue
        
        bonus_data = {'lock_permissions': lock_permissions}
        if parent is None:
            parent_id = None
        else:
            parent_id = parent.id
        bonus_data['parent_id'] = parent_id
        
        data = []
        channels = guild.channels
        for position, channel_id in ordered:
            channel_ = channels[channel_id]
            
            if channel is channel_:
                data.append({'id': channel_id, 'position': position, **bonus_data})
                continue
            
            if channel_.position != position:
                data.append({'id': channel_id, 'position': position})
        
        await self.http.channel_move(guild.id, data, reason)
    
    
    async def channel_edit(
        self, channel, channel_template = None, *, reason = None, **keyword_parameters,
    ):
        """
        Edits the given guild channel. Different channel types accept different fields, so make sure to not pass
        out of place parameters. Only the given fields will be modified of the channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to edit.
        
        channel_template : `None`, ``Channel`` = `None`, Optional
            A channel to use as a template.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        bitrate : `int`, Optional (Keyword only)
            The bitrate (in bits) of the voice channel.
        
        default_forum_layout : ``ForumLayout``, `int`, Optional (Keyword only)
            The default layout used to display threads of the forum.
        
        default_sort_order : ``SortOrder``, `int`, Optional (Keyword only)
            The default thread ordering of the forum.
        
        default_thread_auto_archive_after : `int`, Optional (Keyword only)
            The default duration (in seconds) for newly created threads to automatically archive the themselves.
        
        default_thread_reaction : `None`, ``Emoji``, Optional (Keyword only)
            The emoji to show in the add reaction button on a thread of the forum channel.
                
        default_thread_slowmode : `int`, Optional (Keyword only)
            The default slowmode applied to the channel's threads.
        
        flags : `int`, ``ChannelFlag``, Optional (Keyword only)
            The channel's flags.
        
        name : `str`, Optional (Keyword only)
            The channel's name.
        
        nsfw : `int`, Optional (Keyword only)
            Whether the channel is marked as non safe for work.
        
        parent_id : `None`, `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        
        permission_overwrites : `None`, list` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        
        position : `int`, Optional (Keyword only)
            The channel's position.
        
        region : `None`, ``VoiceRegion``, `str`, Optional (Keyword only)
            The channel's voice region.
        
        slowmode : `int`, Optional (Keyword only)
            The channel's slowmode.
        
        topic : `None`, `str`, Optional (Keyword only)
            The channel's topic.
        
        users : `iterable` of (`int`, ``ClientUserBase``), Optional (Keyword only)
            The users in the channel.
            
        video_quality_mode : ``VideoQualityMode``, Optional (Keyword only)
            The video quality of the voice channel.
        
        Raises
        ------
        TypeError
            - If any parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel, channel_id = get_channel_and_id(channel, Channel.is_in_group_guild)
        data = build_edit_payload(channel, channel_template, CHANNEL_GUILD_FIELD_CONVERTERS, keyword_parameters)
        
        if data:
            await self.http.channel_edit(channel_id, data, reason)
    
    
    async def channel_create(
        self, 
        guild,
        channel_template = None,
        type_ = ...,
        *,
        channel_type = ChannelType.guild_text,
        reason = None,
        **keyword_parameters,
    ):
        """
        Creates a new channel at the given `guild`. If the channel is successfully created returns it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild where the channel will be created.
        
        channel_template : `None`, ``Channel`` = `None`, Optional
            Channel entity to use as a template.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the `guild`'s audit logs.
        
        channel_type : ``ChannelType``, `int` = ``ChannelType.guild_text``, Optional (Keyword only)
            The type of the created channel.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to create the channel with.
        
        Other Parameters
        ----------------
        bitrate : `int`, Optional (Keyword only)
            The bitrate (in bits) of the voice channel.
        
        default_forum_layout : ``ForumLayout``, `int`, Optional (Keyword only)
            The default layout used to display threads of the forum.
        
        default_sort_order : ``SortOrder``, `int`, Optional (Keyword only)
            The default thread ordering of the forum.
        
        default_thread_auto_archive_after : `int`, Optional (Keyword only)
            The default duration (in seconds) for newly created threads to automatically archive the themselves.
        
        default_thread_reaction : `None`, ``Emoji``, Optional (Keyword only)
            The emoji to show in the add reaction button on a thread of the forum channel.
                
        default_thread_slowmode : `int`, Optional (Keyword only)
            The default slowmode applied to the channel's threads.
        
        flags : `int`, ``ChannelFlag``, Optional (Keyword only)
            The channel's flags.
        
        name : `str`, Optional (Keyword only)
            The channel's name.
        
        nsfw : `int`, Optional (Keyword only)
            Whether the channel is marked as non safe for work.
        
        parent_id : `None`, `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        
        permission_overwrites : `None`, list` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        
        position : `int`, Optional (Keyword only)
            The channel's position.
        
        region : `None`, ``VoiceRegion``, `str`, Optional (Keyword only)
            The channel's voice region.
        
        slowmode : `int`, Optional (Keyword only)
            The channel's slowmode.
        
        topic : `None`, `str`, Optional (Keyword only)
            The channel's topic.
        
        users : `iterable` of (`int`, ``ClientUserBase``), Optional (Keyword only)
            The users in the channel.
            
        video_quality_mode : ``VideoQualityMode``, Optional (Keyword only)
            The video quality of the voice channel.
        
        Returns
        -------
        channel : ``Channel``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        # Checkout type
        if type_ is not ...:
            warnings.warn(
                (
                    f'`type_` parameter of `{self.__class__.__name__}.channel_create` is deprecated and will be '
                    f'removed in 2023 February. Please use `channel_type` instead.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            channel_type = type_
        
        keyword_parameters['channel_type'] = channel_type
        
        # Checkout name
        if (channel_template is not None) and isinstance(channel_template, str) and ('name' not in keyword_parameters):
            warnings.warn(
                (
                    f'`name` parameter of `{self.__class__.__name__}.channel_create` is moved to be a keyword only '
                    f'parameter and the positional usage is deprecated and will be removed in 2023 February.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            keyword_parameters['name'] = channel_template
            channel_template = None
        
        data = build_create_payload(channel_template, CHANNEL_GUILD_MAIN_FIELD_CONVERTERS, keyword_parameters)
        channel_data = await self.http.channel_create(guild_id, data, reason)
        
        return Channel.from_data(channel_data, self, guild_id)
    
    
    async def channel_delete(self, channel, *, reason = None):
        """
        Deletes the specified guild channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to delete.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If the given `channel` is not ``Channel``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If a category channel is deleted, it's sub-channels will not be removed, instead they will move under the guild.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_guild)
        
        await self.http.channel_delete(channel_id, reason)
    
    
    async def channel_follow(self, source_channel, target_channel):
        """
        Follows the `source_channel` with the `target_channel`. Returns the webhook, what will crosspost the published
        messages.
        
        This method is a coroutine.
        
        Parameters
        ----------
        source_channel : ``Channel``, `int`
            The channel what will be followed. Must be an announcements channel.
        target_channel : ``Channel``, `int`instance
            The target channel where the webhook messages will be sent. Can be any guild text channel type.
        
        Returns
        -------
        webhook : ``Webhook``
            The webhook what will crosspost the published messages. This webhook has no `.token` set.
        
        Raises
        ------
        TypeError
            - If the `source_channel` was not given neither as ``Channel`` nor `int`.
            - If the `target_channel` was not given neither as ``Channel`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        source_channel, source_channel_id = get_channel_and_id(source_channel, Channel.is_guild_announcements)
        if source_channel is None:
            source_channel = create_partial_channel_from_id(source_channel_id, ChannelType.guild_announcements, 0)
        
        target_channel_id = get_channel_id(target_channel, Channel.is_in_group_guild_system)
        
        data = {
            'webhook_channel_id': target_channel_id,
        }
        
        data = await self.http.channel_follow(source_channel_id, data)
        webhook = await Webhook._from_follow_data(data, source_channel, target_channel_id, self)
        return webhook
    

    async def permission_overwrite_edit(self, channel, permission_overwrite, *, reason = None, **keyword_parameters):
        """
        Edits the given permission overwrite.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ˙˙Channel``, `int`
            The channel where the permission overwrite is.
        
        permission_overwrite : ``PermissionOverwrite``
            The permission overwrite to edit.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        allow : `None`, ``Permission``, `int`, Optional (Keyword only)
            The permission overwrite's allowed permission's value.
        
        deny : `None`, ``Permission``, `int`, Optional (Keyword only)
            The permission overwrite's denied permission's value.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_guild_sortable)
        target_id = get_permission_overwrite_target_id(permission_overwrite)
        
        data = build_create_payload(
            permission_overwrite, PERMISSION_OVERWRITE_PERMISSION_FIELD_CONVERTERS, keyword_parameters
        )
        
        await self.http.permission_overwrite_create(channel_id, target_id, data, reason)
    
    
    async def permission_overwrite_delete(self, channel, permission_overwrite, *, reason = None):
        """
        Deletes the given permission overwrite.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ˙˙Channel``
            The channel where the permission overwrite is.
        
        permission_overwrite : ``PermissionOverwrite``, `int`
            The permission overwrite to delete.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.

        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_guild_sortable)
        target_id = get_permission_overwrite_target_id(permission_overwrite)
        
        await self.http.permission_overwrite_delete(channel_id, target_id, reason)
    
    
    async def permission_overwrite_create(
        self, channel, permission_overwrite_template = None, *, reason = None, **keyword_parameters
    ):
        """
        Creates a permission overwrite at the given channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to what the permission overwrite will be added.
        
        permission_overwrite_template : `None`, ``PermissionOverwrite`` = `None`, Optional
            Permission overwrite to be used as a template for creating the new one.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        allow : `None`, ``Permission``, `int`, Optional (Keyword only)
            The permission overwrite's allowed permission's value.
        
        deny : `None`, ``Permission``, `int`, Optional (Keyword only)
            The permission overwrite's denied permission's value.
        
        target : `int`, ``Role``, ``ClientUserBase``, Optional (Keyword only)
            The permission overwrite's target. Shortcut for defining `target_id` and `target_type` with 1 parameter.
        
        target_id : `int`, Optional (Keyword only)
            The permission overwrite's target's identifier.
        
        target_type : `None`, ``PermissionOverwriteTargetType``, Optional (Keyword only)
            The permission overwrite's target's type. Required if `target_id` is given as a snowflake.
        
        
        Returns
        -------
        permission_overwrite : ``PermissionOverwrite``
            A permission overwrite, what estimatedly is same as the one what Discord will create.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_guild_sortable)
        
        if (
            (permission_overwrite_template is not None) and
            (not isinstance(permission_overwrite_template, PermissionOverwrite)) and
            ('target' not in keyword_parameters)
        ):
            warnings.warn(
                (
                    f'`target` parameter of `{self.__class__.__name__}.permission_overwrite_create` is moved to be '
                    f'a keyword only parameter and the positional usage is deprecated and will be removed in 2023 '
                    f'February.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            keyword_parameters['target'] = permission_overwrite_template
        
        data = build_create_payload(
            permission_overwrite_template, PERMISSION_OVERWRITE_FIELD_CONVERTERS, keyword_parameters
        )
        
        try:
            target_id = data['id']
        except KeyError:
            raise RuntimeError(
                f'Cannot create permission overwrite to unknown target. Parameters are already destructed, no '
                f'additional context is providable.'
            )
        
        await self.http.permission_overwrite_create(channel_id, target_id, data, reason)
        return PermissionOverwrite.from_data(data)
    
    
    async def guild_channel_get_all(self, guild):
        """
        Requests the given guild's channels and if there any de-sync between the wrapper and Discord, applies the
        changes.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's channels will be requested.
        
        Returns
        -------
        channels : `list` of ``Channel``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild``, nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        data = await self.http.guild_channel_get_all(guild_id)
        if guild is None:
            guild = create_partial_guild_from_id(guild_id)
        
        guild._sync_channels(data)
        
        return [*guild.channels.values()]
    
    
    async def forum_tag_create(self, forum_channel, forum_tag = None, *, reason = None, **keyword_parameters):
        """
        Creates a new forum tag in the channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        forum_channel : ``Channel``, `int`
            The channel to create the tag in.
        
        forum_tag : ``ForumTag``, `None` = `None`, Optional
            A forum tag which can be used as a template for the newly created tag.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            The tag's emoji.
            
        moderated : `bool`, Optional (Keyword only)
            Whether this tag can only be added or removed by a user with `manage_threads` permission.
        
        name : `str`, Optional (Keyword only)
            The tag's name.
        
        Returns
        -------
        forum_tag : `None`, ``ForumTag``
            The created forum tag.
        
        Raises
        ------
        TypeError
            If a parameter's type is not acceptable.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        A forum channel can have up to 25 tags.
        
        See Also
        --------
        - ``.forum_tag_delete`` : Delete a forum tag.
        - ``.forum_tag_edit`` : Modify a forum tag.
        """
        channel_id = get_channel_id(forum_channel, Channel.is_guild_forum)
        data = build_create_payload(forum_tag, FORUM_TAG_FIELD_CONVERTERS, keyword_parameters)
        
        channel_data = await self.http.forum_tag_create(channel_id, data, reason)
        
        # Fixing discord bug: Returns channel data instead of forum tag.
        available_tag_data_array = channel_data.get('available_tags', None)
        if (available_tag_data_array is None) or (not available_tag_data_array):
            # Cannot build forum tag from nothing, eww..
            return None
        
        available_tag_data_array = sorted(
            available_tag_data_array, key = _forum_tag_data_array_sort_key, reverse = True
        )
        
        for forum_tag_data in available_tag_data_array:
            for key, value in data.items():
                if forum_tag_data.get(key, None) != value:
                    break
            
            else:
                # All fields matched.
                break
            
            continue
        
        else:
            # If no fields matched, return the newest tag.
            forum_tag_data = available_tag_data_array[0]
        
        return ForumTag.from_data(forum_tag_data)
    
    
    async def forum_tag_edit(
        self, forum_channel, forum_tag, template_forum_tag = None, *, reason = None, **keyword_parameters
    ):
        """
        Edits the given forum tag.
        
        This method is a coroutine.
        
        Parameters
        ----------
        forum_channel : ``Channel``, `int`
            The channel to edit the tag in.
        
        forum_tag : ``ForumTag``, `int`
            The forum tag to edit.
        
        template_forum_tag : ``ForumTag``, `None` = `None`, Optional
            A forum tag which can be used as a template for edition.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            The tag's emoji.
            
        moderated : `bool`, Optional (Keyword only)
            Whether this tag can only be added or removed by a user with `manage_threads` permission.
        
        name : `str`, Optional (Keyword only)
            The tag's name.
        
        Raises
        ------
        TypeError
            If a parameter's type is not acceptable.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        - ``.forum_tag_create`` : Create a forum tag.
        - ``.forum_tag_delete`` : Delete a forum tag.
        """
        channel_id = get_channel_id(forum_channel, Channel.is_guild_forum)
        forum_tag, forum_tag_id = get_forum_tag_and_id(forum_tag)
        data = build_edit_payload(forum_tag, template_forum_tag, FORUM_TAG_FIELD_CONVERTERS, keyword_parameters)
        
        if data:
            # Fixing discord bug: name.BASE_TYPE_REQUIRED('This field is required')
            if (forum_tag is not None) and ('name' not in data):
                data['name'] = forum_tag.name
            
            await self.http.forum_tag_edit(channel_id, forum_tag_id, data, reason)
    
    
    async def forum_tag_delete(self, forum_channel, forum_tag, *, reason = None):
        """
        Deletes the given forum tag.
        
        Parameters
        ----------
        forum_channel : ``Channel``, `int`
            The channel to edit the tag in.
        
        forum_tag : ``ForumTag``, `int`
            The forum tag to delete.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If a parameter's type is not acceptable.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        - ``.forum_tag_create`` : Create a forum tag.
        - ``.forum_tag_edit`` : Modify a forum tag.
        """
        channel_id = get_channel_id(forum_channel, Channel.is_guild_forum)
        forum_tag_id = get_forum_tag_id(forum_tag)
        await self.http.forum_tag_delete(channel_id, forum_tag_id, reason)



__all__ = ('AutoModerationAction',)

from scarletio import RichAttributeErrorBaseType, copy_docs

from ..action_metadata import AutoModerationActionMetadataBase

from .fields import parse_metadata, parse_type, put_metadata_into, put_type_into, validate_type
from .helpers import guess_action_type_from_keyword_parameters
from .preinstanced import AutoModerationActionType


class AutoModerationAction(RichAttributeErrorBaseType):
    """
    An action that is executed when a rule is triggered.
    
    Attributes
    ----------
    metadata : ``AutoModerationActionMetadataBase``
        Characterises the action
    type : ``AutoModerationActionType``
        The action's type.
    """
    __slots__ = ('metadata', 'type')
    
    def __new__(cls, action_type = None, **keyword_parameters):
        """
        Creates a new auto moderation action from the given parameters.
        
        Parameters
        ----------
        action_type : ``AutoModerationActionType``, `int`, Optional
            The moderation action's type.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to pass
        
        Other Parameters
        ----------------
        channel_id : `None`, ``Channel``, `int`, Optional (Keyword only)
            The channel where the alert messages should be sent.
            
            > Mutually exclusive with the `duration` parameter.
        
        duration : `None`, `int`, `float`, Optional (Keyword only)
            The timeout's duration applied on trigger.
            
            > Mutually exclusive with the `channel` parameter.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Parameter mismatch.
        """
        action_type = validate_type(action_type)
        action_type = guess_action_type_from_keyword_parameters(action_type, keyword_parameters)
        metadata = action_type.metadata_type(**keyword_parameters)
        
        self = object.__new__(cls)
        self.type = action_type
        self.metadata = metadata
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``AutoModerationAction`` from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received auto moderation action data.
        
        Returns
        -------
        self : ``AutoModerationAction``
            The created auto moderation rule.
        """
        action_type = parse_type(data)
        metadata = parse_metadata(data, action_type)
        
        self = object.__new__(cls)
        self.type = action_type
        self.metadata = metadata
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the auto moderation action to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        put_type_into(self.type, data, defaults)
        put_metadata_into(self.metadata, data, defaults)
        return data
    
    
    def __eq__(self, other):
        """Returns whether the two auto moderation actions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.type is not other.type:
            return False
        
        if self.metadata != other.metadata:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the auto moderation action's hash value."""
        hash_value = 0
        
        hash_value ^= self.type.value
        
        metadata = self.metadata
        if (metadata is not None):
            hash_value ^= hash(metadata)
        
        return hash_value
    
    
    def __repr__(self):
        """Returns the auto moderation action's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # type
        type_ = self.type
        
        repr_parts.append(' type = ')
        repr_parts.append(repr(type_.name))
        repr_parts.append('~')
        repr_parts.append(repr(type_.value))
        
        # metadata
        metadata = self.metadata
        if (metadata is not None):
            repr_parts.append(', metadata = ')
            repr_parts.append(repr(metadata))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def copy(self):
        """
        Copies the auto moderation action.
        
        Returns
        -------
        new : ``AutoModerationAction``
        """
        new = object.__new__(type(self))
        
        # metadata
        new.metadata = self.metadata.copy()
        
        # type
        new.type = self.type
        
        return new

    
    def copy_with(self, *, action_type = ..., **keyword_parameters):
        """
        Copies the auto moderation action with the given attributes replaced.
        
        Parameters
        ----------
        action_type : ``AutoModerationActionType``, `int`, Optional (Keyword only)
            The moderation action's type.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to pass
        
        Other Parameters
        ----------------
        channel_id : `None`, ``Channel``, `int`, Optional (Keyword only)
            The channel where the alert messages should be sent.
            
            > Mutually exclusive with the `duration` parameter.
        
        duration : `None`, `int`, `float`, Optional (Keyword only)
            The timeout's duration applied on trigger.
            
            > Mutually exclusive with the `channel` parameter.
        
        Returns
        -------
        self : ``AutoModerationAction``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        """
        if action_type is ...:
            if keyword_parameters:
                action_type = guess_action_type_from_keyword_parameters(AutoModerationActionType.none, keyword_parameters)
            else:
                action_type = self.type
        else:
            action_type = validate_type(action_type)
        
        if action_type is self.type:
            metadata = self.metadata.copy_with(**keyword_parameters)
        else:
            metadata = action_type.metadata_type(**keyword_parameters)
        
        new = object.__new__(type(self))
        new.type = action_type
        new.metadata = metadata
        return new
    
    
    # ---- Field proxies ----
    
    @property
    @copy_docs(AutoModerationActionMetadataBase.channel_id)
    def channel_id(self):
        return self.metadata.channel_id
    
    
    @property
    @copy_docs(AutoModerationActionMetadataBase.duration)
    def duration(self):
        return self.metadata.duration
    
    # ---- Utility proxies ----
    
    @property
    @copy_docs(AutoModerationActionMetadataBase.channel)
    def channel(self):
        return self.metadata.channel



__all__ = ()

from scarletio import export

from .....discord.application_command import ApplicationCommandOptionType


APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND = ApplicationCommandOptionType.sub_command
APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND_CATEGORY = ApplicationCommandOptionType.sub_command_group

APPLICATION_COMMAND_FUNCTION_DEEPNESS = -1
APPLICATION_COMMAND_HANDLER_DEEPNESS = 0
APPLICATION_COMMAND_DEEPNESS = 1
APPLICATION_COMMAND_CATEGORY_DEEPNESS_START = 2
APPLICATION_COMMAND_CATEGORY_DEEPNESS_MAX = 3


export(APPLICATION_COMMAND_FUNCTION_DEEPNESS, 'APPLICATION_COMMAND_FUNCTION_DEEPNESS')



__all__ = ('RoleManagerType',)

from ...bases import Preinstance as P, PreinstancedBase

from ..role_manager_metadata import (
    RoleManagerMetadataBase, RoleManagerMetadataBooster, RoleManagerMetadataBot, RoleManagerMetadataIntegration,
    RoleManagerMetadataSubscription
)

class RoleManagerType(PreinstancedBase):
    """
    Represents a managed role's manager type.
    
    Attributes
    ----------
    name : `str`
        The name of the role manager type.
    value : `int`
        The identifier value the role manager type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``RoleManagerType``) items
        Stores the predefined ``RoleManagerType``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The role manager types' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the role manager types.
    
    Every predefined role manager type can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+---------------------------------------+
    | Class attribute name  | Name          | Value | Metadata type                         |
    +=======================+===============+=======+=======================================+
    | none                  | none          | 0     | ``RoleManagerMetadataBase``           |
    +-----------------------+---------------+-------+---------------------------------------+
    | unset                 | unset         | 1     | ``RoleManagerMetadataBase``           |
    +-----------------------+---------------+-------+---------------------------------------+
    | unknown               | unknown       | 2     | ``RoleManagerMetadataBase``           |
    +-----------------------+---------------+-------+---------------------------------------+
    | bot                   | bot           | 3     | ``RoleManagerMetadataBot``            |
    +-----------------------+---------------+-------+---------------------------------------+
    | booster               | booster       | 4     | ``RoleManagerMetadataBooster``        |
    +-----------------------+---------------+-------+---------------------------------------+
    | integration           | integration   | 5     | ``RoleManagerMetadataIntegration``    |
    +-----------------------+---------------+-------+---------------------------------------+
    | subscription          | subscription  | 6     | ``RoleManagerMetadataSubscription``   |
    +-----------------------+---------------+-------+---------------------------------------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ('metadata_type', )
    
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new role manager type with the given value.
        
        Parameters
        ----------
        value : `int`
            Value representing the role manager.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.name = value
        self.value = value
        self.metadata_type = RoleManagerMetadataBase
        return self
    
    
    def __init__(self, value, name, metadata_type):
        """
        Creates a new scheduled event entity type instance from the given parameters.
        
        Parameters
        ----------
        value : `str`
            The unique identifier of the scheduled event entity type.
        name : `str`
            The name of the scheduled event entity type.
        metadata_type : `None`, ``RoleManagerMetadataBase`` subclass
            The role manager's metadata's type.
        """
        self.name = name
        self.value = value
        self.metadata_type = metadata_type
        self.INSTANCES[value] = self
    
    
    def __bool__(self):
        """Returns whether the role manager's type is set."""
        if self.value:
            boolean = True
        else:
            boolean = False
        
        return boolean
    
    
    none = P(0, 'none', RoleManagerMetadataBase)
    unset = P(1, 'unset', RoleManagerMetadataBase)
    unknown = P(2, 'unknown', RoleManagerMetadataBase)
    bot = P(3, 'bot', RoleManagerMetadataBot)
    booster = P(4, 'booster', RoleManagerMetadataBooster)
    integration = P(5, 'integration', RoleManagerMetadataIntegration)
    subscription = P(6, 'subscription', RoleManagerMetadataSubscription)



from .achievement import *
from .application_command import *
from .application_role_connection import *
from .auto_moderation import *
from .channel import *
from .client import *
from .client_gayteway import *
from .discovery import *
from .emoji import *
from .guild import *
from .guild_ban import *
from .interaction import *
from .integration import *
from .invite import *
from .locked import *
from .message import *
from .oauth2 import *
from .miscellaneous import *
from .reaction import *
from .role import *
from .scheduled_event import *
from .stage import *
from .sticker import *
from .thread import *
from .user import *
from .webhook import *


__all__ = (
    *achievement.__all__,
    *application_command.__all__,
    *application_role_connection.__all__,
    *auto_moderation.__all__,
    *channel.__all__,
    *client.__all__,
    *client_gayteway.__all__,
    *discovery.__all__,
    *emoji.__all__,
    *guild.__all__,
    *guild_ban.__all__,
    *interaction.__all__,
    *integration.__all__,
    *invite.__all__,
    *locked.__all__,
    *message.__all__,
    *miscellaneous.__all__,
    *oauth2.__all__,
    *reaction.__all__,
    *role.__all__,
    *scheduled_event.__all__,
    *stage.__all__,
    *sticker.__all__,
    *thread.__all__,
    *user.__all__,
    *webhook.__all__,
)


from .achievement import ClientCompoundAchievementEndpoints
from .application_command import ClientCompoundApplicationCommandEndpoints
from .application_role_connection import ClientCompoundApplicationRoleConnectionEndpoints
from .auto_moderation import ClientCompoundAutoModerationEndpoints
from .channel import ClientCompoundChannelEndpoints
from .client import ClientCompoundClientEndpoints
from .client_gayteway import ClientCompoundClientGateway
from .discovery import ClientCompoundDiscoveryEndpoints
from .emoji import ClientCompoundEmojiEndpoints
from .guild import ClientCompoundGuildEndpoints
from .guild_ban import ClientCompoundGuildBanEndpoints
from .interaction import ClientCompoundInteractionEndpoints
from .integration import ClientCompoundIntegrationEndpoints
from .invite import ClientCompoundInviteEndpoints
from .locked import ClientCompoundLockedEndpoints
from .message import ClientCompoundMessageEndpoints
from .miscellaneous import ClientCompoundMiscellaneousEndpoints
from .oauth2 import ClientCompoundOauth2Endpoints
from .reaction import ClientCompoundReactionEndpoints
from .role import ClientCompoundRoleEndpoints
from .scheduled_event import ClientCompoundScheduledEventEndpoints
from .stage import ClientCompoundStageEndpoints
from .sticker import ClientCompoundStickerEndpoints
from .thread import ClientCompoundThreadEndpoints
from .user import ClientCompoundUserEndpoints
from .webhook import ClientCompoundWebhookEndpoints


CLIENT_COMPOUNDS = (
    ClientCompoundAchievementEndpoints,
    ClientCompoundApplicationCommandEndpoints,
    ClientCompoundApplicationRoleConnectionEndpoints,
    ClientCompoundAutoModerationEndpoints,
    ClientCompoundChannelEndpoints,
    ClientCompoundClientEndpoints,
    ClientCompoundClientGateway,
    ClientCompoundDiscoveryEndpoints,
    ClientCompoundEmojiEndpoints,
    ClientCompoundGuildEndpoints,
    ClientCompoundGuildBanEndpoints,
    ClientCompoundInteractionEndpoints,
    ClientCompoundIntegrationEndpoints,
    ClientCompoundInviteEndpoints,
    ClientCompoundLockedEndpoints,
    ClientCompoundMessageEndpoints,
    ClientCompoundMiscellaneousEndpoints,
    ClientCompoundOauth2Endpoints,
    ClientCompoundReactionEndpoints,
    ClientCompoundRoleEndpoints,
    ClientCompoundScheduledEventEndpoints,
    ClientCompoundStageEndpoints,
    ClientCompoundStickerEndpoints,
    ClientCompoundThreadEndpoints,
    ClientCompoundUserEndpoints,
    ClientCompoundWebhookEndpoints,
)



__all__ = ()

from .....discord.guild import Guild
from .....discord.application_command.application_command import _assert__application_command__nsfw
from .....discord.application_command.constants import (
    APPLICATION_COMMAND_NAME_LENGTH_MAX, APPLICATION_COMMAND_NAME_LENGTH_MIN
)
from .....discord.permission import Permission
from .....discord.preconverters import preconvert_bool, preconvert_flag, preconvert_snowflake

from ...utils import UNLOADING_BEHAVIOUR_DELETE, UNLOADING_BEHAVIOUR_INHERIT, UNLOADING_BEHAVIOUR_KEEP


def _reset_application_command_schema(entity):
    """
    Resets the slasher application commands schema.
    
    Parameters
    ----------
    entity : ``CommandBaseApplicationCommand``
        The command to reset's its schema if applicable.
    """
    schema = entity._schema
    if (schema is not None):
        entity._schema = None
        
        parent_reference = entity._parent_reference
        if (parent_reference is not None):
            slasher = parent_reference()
            
            if (slasher is not None):
                slasher._schema_reset(entity)


def _validate_delete_on_unload(delete_on_unload):
    """
    Validates the given `delete_on_unload` value.
    
    Parameters
    ----------
    delete_on_unload : `None`, `bool`
        The `delete_on_unload` value to validate.
    
    Returns
    -------
    unloading_behaviour : `int`
        The validated `delete_on_unload` value.
    
    Raises
    ------
    TypeError
        If `delete_on_unload` was not given as `None` nor as `bool`.
    """
    if delete_on_unload is None:
        unloading_behaviour = UNLOADING_BEHAVIOUR_INHERIT
    else:
        delete_on_unload = preconvert_bool(delete_on_unload, 'delete_on_unload')
        if delete_on_unload:
            unloading_behaviour = UNLOADING_BEHAVIOUR_DELETE
        else:
            unloading_behaviour = UNLOADING_BEHAVIOUR_KEEP
    
    return unloading_behaviour


def _validate_allow_by_default(allow_by_default):
    """
    Validates the given `allow_by_default` value.
    
    Parameters
    ----------
    allow_by_default : `None`, `bool`
        The `allow_by_default` value to validate.
    
    Returns
    -------
    allow_by_default : `None`, `bool`
        The validated `allow_by_default` value.
    
    Raises
    ------
    TypeError
        If `allow_by_default` was not given as `None`, `bool`.
    """
    if (allow_by_default is not None):
        allow_by_default = preconvert_bool(allow_by_default, 'allow_by_default')
    
    return allow_by_default


def _validate_allow_in_dm(allow_in_dm):
    """
    Validates the given `allow_in_dm` value.
    
    Parameters
    ----------
    allow_in_dm : `None`, `bool`
        The `allow_in_dm` value to validate.
    
    Returns
    -------
    allow_in_dm : `None`, `bool`
        The validated `allow_in_dm` value.
    
    Raises
    ------
    TypeError
        If `allow_in_dm` was not given as `None`, `bool`.
    """
    if (allow_in_dm is not None):
        allow_in_dm = preconvert_bool(allow_in_dm, 'allow_in_dm')
    
    return allow_in_dm


def _validate_is_global(is_global):
    """
    Validates the given `is_global` value.
    
    Parameters
    ----------
    is_global : `None`, `bool`
        The `is_global` value to validate.
    
    Returns
    -------
    is_global : `bool`
        The validated `is_global` value.
    
    Raises
    ------
    TypeError
        If `is_global` was not given as `None` nor as `bool`.
    """
    if is_global is None:
        is_global = False
    else:
        is_global = preconvert_bool(is_global, 'is_global')
    
    return is_global


def _validate_1_guild(guild):
    """
    Validates 1 guild value.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The guild value to validate.
    
    Returns
    -------
    guild_id : `int`
        Validated guild value converted to `int`.
    
    Raises
    ------
    TypeError
        If `guild` was not given neither as ``Guild`` nor `int`.
    ValueError
        If `guild` is an integer out of uint64 value range.
    """
    if isinstance(guild, Guild):
        guild_id = guild.id
    elif isinstance(guild, (int, str)):
        guild_id = preconvert_snowflake(guild, 'guild')
    else:
        raise TypeError(
            f'`guild`can be `{Guild.__class__.__name__}`, `int`, got {guild.__class__.__name__}; {guild!r}.'
        )
    
    return guild_id


def _validate_guild(guild):
    """
    Validates the given `guild` value.
    
    Parameters
    ----------
    guild : `None`, `int`, ``Guild``, (`list`, `set`) of (`int`, ``Guild``
        The `is_global` value to validate.
    
    Returns
    -------
    guild_ids : `None`, `set` of `int`
        The validated `guild` value.
    
    Raises
    ------
    TypeError
        If `guild` was not given neither as `None`, ``Guild``,  `int`, (`list`, `set`) of (`int`, ``Guild``)
    ValueError
        - If `guild` is given as an empty container.
        - If `guild` is or contains an integer out of uint64 value range.
    """
    if guild is None:
        guild_ids = None
    else:
        guild_ids = set()
        if isinstance(guild, (list, set)):
            for guild_value in guild:
                guild_id = _validate_1_guild(guild_value)
                guild_ids.add(guild_id)
        else:
            guild_id = _validate_1_guild(guild)
            guild_ids.add(guild_id)
        
        if not guild_ids:
            raise ValueError(
                f'`guild` cannot be empty container, got {guild!r}.'
            )
    
    return guild_ids


def _validate_name(name):
    """
    Validates the given name.
    
    Parameters
    ----------
    name : `None`, `str`
        A command's respective name.
    
    Returns
    -------
    name : `None`, `str`
        The validated name.
    
    Raises
    ------
    TypeError
        If `name` is not given as `None` neither as `str`.
    ValueError
        If `name` length is out of the expected range [1:32].
    """
    if name is not None:
        name_type = name.__class__
        if name_type is str:
            pass
        elif issubclass(name_type, str):
            name = str(name)
        else:
            raise TypeError(
                f'`name` can be `None`, `str`, got {name_type.__name__}; {name!r}.'
            )
        
        name_length = len(name)
        if (
            name_length < APPLICATION_COMMAND_NAME_LENGTH_MIN or
            name_length > APPLICATION_COMMAND_NAME_LENGTH_MAX
        ):
            raise ValueError(
                f'`name` length is out of the expected range '
                f'[{APPLICATION_COMMAND_NAME_LENGTH_MIN}:'
                f'{APPLICATION_COMMAND_NAME_LENGTH_MAX}], got {name_length!r}; {name!r}.'
            )
    
    return name


def _validate_nsfw(nsfw):
    """
    Validates the given `nsfw` value.
    
    Parameters
    ----------
    nsfw : `None`, `bool`
        The `nsfw` value to validate.
    
    Returns
    -------
    nsfw : `None`, `bool`
        The validated `nsfw` value.
    
    Raises
    ------
    TypeError
        If `nsfw` was not given as `None`, `bool`.
    """
    try:
        _assert__application_command__nsfw(nsfw)
    except AssertionError as err:
        raise TypeError(*err.args) from err
    
    return nsfw


def _validate_required_permissions(required_permissions):
    """
    Validates the given `required_permissions` value.
    
    Parameters
    ----------
    required_permissions : `None`, `int`, ``Permission``
        The `required_permissions` value to validate.
    
    Returns
    -------
    required_permissions : `None`, ``Permission``
        The validated `required_permissions` value.
    
    Raises
    ------
    TypeError
        If `required_permissions` was not given as `None`, ``Permission``, `int`.
    """
    if (required_permissions is not None):
        required_permissions = preconvert_flag(required_permissions, 'required_permissions', Permission)
    
    return required_permissions



__all__ = ('ScheduledEventEntityMetadataBase',)

from scarletio import RichAttributeErrorBaseType


class ScheduledEventEntityMetadataBase(RichAttributeErrorBaseType):
    """
    Base class for ``ScheduledEvent``'s entity metadata.
    """
    __slots__ = ()
    
    def __new__(cls, keyword_parameters):
        """
        Creates a new entity metadata instance.
        """
        raise NotImplementedError
        
    def __repr__(self):
        """Returns the entity metadata's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new scheduled event entity metadata instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Entity metadata structure.
        
        Returns
        -------
        self : `instance<cls>`
        """
        raise NotImplementedError
    
    
    def to_data(self):
        """
        Converts the entity metadata to json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    
    def __eq__(self, other):
        """Returns whether the two entity metadatas equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return True
    
    
    def __hash__(self):
        """Returns the entity metadata's hash value."""
        return 0



__all__ = ()

from math import floor, log10

from .... import __package__ as PACKAGE_NAME
from ... import REGISTERED_COMMANDS, REGISTERED_COMMANDS_BY_NAME, command_sort_key, normalize_command_name, register
from ...core.helpers import render_main_call_into


PACKAGE = __import__(PACKAGE_NAME)

BREAK_LINE = '─' * 80


def show_command(command_name, sub_command_names):
    command = REGISTERED_COMMANDS_BY_NAME.get(normalize_command_name(command_name), None)
    
    output_parts = []
    if command is None:
        output_parts = []
        
        output_parts.append('No command is added for: ')
        output_parts.append(repr(command_name))
        output_parts.append('\n\n')
        output_parts.append('Try using "$ ')
        render_main_call_into(output_parts)
        output_parts.append(' help" to list all available commands.\n')
    
    else:
        command.render_direct_usage_into(output_parts, *sub_command_names)
        output_parts.append('\n')
    
    return ''.join(output_parts)


def list_commands():
    output_parts = []
    output_parts.append('Available commands:\n\n')
    
    command_count = len(REGISTERED_COMMANDS)
    index_adjust = floor(log10(command_count)) + 1
    
    for index, command in enumerate(sorted(REGISTERED_COMMANDS, key = command_sort_key), 1):
        command_name = command.name
        output_parts.append(str(index).rjust(index_adjust))
        output_parts.append('. ')
        output_parts.append(command_name)
        output_parts.append('\n')
    
    output_parts.append('\n')
    output_parts.append('Use "$ ')
    render_main_call_into(output_parts)
    output_parts.append(' help ')
    output_parts.append('COMMAND-NAME')
    output_parts.append('" for more information.\n')
    
    return ''.join(output_parts)


@register(
    name = 'help',
    alters = 'h'
)
def help_(command_name: str = None, *sub_command_names):
    """Either lists the available commands, or shows the command's usage."""
    if command_name is None:
        return list_commands()
    else:
        return show_command(command_name, sub_command_names)


@register(
    into = help_
)
def list_all():
    """Shows the help for all the available commands with their sub-commands included."""
    output_parts = []
    field_added = False
    
    for command in sorted(REGISTERED_COMMANDS, key = command_sort_key):
        for output_parts in command.walk_usage_into(output_parts):
            output_parts.append('\n\n')
            output_parts.append(BREAK_LINE)
            output_parts.append('\n\n')
            field_added = True
    
    if field_added:
        del output_parts[-3:]
    
    return ''.join(output_parts)



from .constants import *
from .fields import *
from .interaction_event import InteractionEvent
from .preinstanced import *


__all__ = (
    *constants.__all__,
    *fields.__all__,
    *interaction_event.__all__,
    *preinstanced.__all__,
)



__all__ = ('InviteTargetType', 'InviteType')

from ..bases import Preinstance as P, PreinstancedBase


class InviteTargetType(PreinstancedBase):
    """
    Represents an ``Invite``'s target's type.
    
    Attributes
    ----------
    name : `str`
        The name of the target type.
    value : `int`
        The Discord side identifier value of the target type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``InviteTargetType``) items
        Stores the predefined ``InviteTargetType``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The invite target types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the invite target types.
    
    Every predefined invite target type can be accessed as class attribute as well:
    
    +-------------------------------+-------------------------------+-------+
    | Class attribute name          | name                          | value |
    +===============================+===============================+=======+
    | none                          | none                          | 0     |
    +-------------------------------+-------------------------------+-------+
    | stream                        | stream                        | 1     |
    +-------------------------------+-------------------------------+-------+
    | embedded_application          | embedded application          | 2     |
    +-------------------------------+-------------------------------+-------+
    | role_subscription_purchase    | role subscription purchase    | 2     |
    +-------------------------------+-------------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    stream = P(1, 'stream')
    embedded_application = P(2, 'embedded application')
    role_subscription_purchase = P(3, 'role subscription purchase')


class InviteType(PreinstancedBase):
    """
    Represents an ``Invite``'s type.
    
    Attributes
    ----------
    name : `str`
        The name of the invite type.
    value : `int`
        The Discord side identifier value of the invite type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``InviteType``) items
        Stores the predefined ``InviteType``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The invite types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the invite types.
    
    Every predefined invite type can be accessed as class attribute as well:
    
    +-----------------------+-----------------------+-------+
    | Class attribute name  | name                  | value |
    +=======================+=======================+=======+
    | guild                 | guild                 | 0     |
    +-----------------------+-----------------------+-------+
    | group_channel         | group channel         | 1     |
    +-----------------------+-----------------------+-------+
    | friend                | friend                | 2     |
    +-----------------------+-----------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    guild = P(0, 'guild')
    group_channel = P(1, 'group channel')
    friend = P(2, 'friend')



from .base import *
from .booster import *
from .bot import *
from .constants import *
from .fields import *
from .integration import *
from .subscription import *


__all__ = (
    *base.__all__,
    *booster.__all__,
    *bot.__all__,
    *constants.__all__,
    *fields.__all__,
    *integration.__all__,
    *subscription.__all__,
)



__all__ = ('AutoModerationRuleTriggerMetadataBase', )

from scarletio import RichAttributeErrorBaseType

from ...bases import PlaceHolder


class AutoModerationRuleTriggerMetadataBase(RichAttributeErrorBaseType):
    """
    Base class for ``AutoModerationRule``'s trigger metadata.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new trigger metadata instance.
        """
        return object.__new__(cls)
        
    def __repr__(self):
        """Returns the trigger metadata's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new auto moderation rule trigger metadata instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Auto moderation rule trigger metadata payload.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the trigger metadata to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    
    def __eq__(self, other):
        """Returns whether the two trigger metadatas are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return True
    
    
    def __hash__(self):
        """Returns the trigger metadata's hash value."""
        return 0
    
    
    def copy(self):
        """
        Copies the trigger metadata.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return object.__new__(type(self))
    
    
    def copy_with(self):
        """
        Copies the trigger metadata with altering it's attributes based on the given fields.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return self.copy()
    
    
    # ---- Place holders ----
    
    excluded_keywords = PlaceHolder(
        None,
        """
        Excluded keywords from under the rule.
        
        Returns
        -------
        excluded_keywords : `None`, `tuple` of `str`
        """
    )
    
    
    keyword_presets = PlaceHolder(
        None,
        """
        Keyword preset defined by Discord which will be searched for in content.
        
        Returns
        -------
        keyword_presets : `None`, `tuple` of ``AutoModerationKeywordPresetType``
        """
    )
    
    
    keywords = PlaceHolder(
        None,
        """
        Substrings which will be searched for in content.
        
        Returns
        -------
        keywords : `None`, `tuple` of `str`
        """
    )
    
    
    regex_patterns = PlaceHolder(
        None,
        """
        Regular expression patterns which are matched against content.
        
        Returns
        -------
        regex_patterns : `None`, `tuple` of `str`
        """
    )
    
    
    mention_limit = PlaceHolder(
        0,
        """
        The amount of mentions in a message after the rule is triggered.
        
        Returns
        -------
        mention_limit : `int`
        """
    )
    
    # ---- Extra utility ----
    
    def iter_excluded_keywords(self):
        """
        Iterates over the excluded keyword of the keyword trigger metadata.
        
        This method is an iterable generator.
        
        Yields
        ------
        excluded_keyword : `str`
        """
        excluded_keywords = self.excluded_keywords
        if (excluded_keywords is not None):
            yield from excluded_keywords
    
    
    def iter_keyword_presets(self):
        """
        Iterates over the keyword presets of the keyword trigger metadata.
        
        This method is an iterable generator.
        
        Yields
        ------
        keyword_preset : `AutoModerationKeywordPresetType`
        """
        keyword_presets = self.keyword_presets
        if (keyword_presets is not None):
            yield from keyword_presets

    
    def iter_keywords(self):
        """
        Iterates over the keywords of the keyword trigger metadata.
        
        This method is an iterable generator.
        
        Yields
        ------
        keyword : `str`
        """
        keywords = self.keywords
        if (keywords is not None):
            yield from keywords
    
    
    def iter_regex_patterns(self):
        """
        Iterates over the regex patterns of the keyword trigger metadata.
        
        This method is an iterable generator.
        
        Yields
        ------
        regex_patterns : `str`
        """
        regex_patterns = self.regex_patterns
        if (regex_patterns is not None):
            yield from regex_patterns



"""
To set a html class to a token, do:

```py
set_highlight_html_class(token_type_identifier, html_class_name)
```

Testing
-------
To check highlights set some colors down and enjoy.

```py
TUMMY_REPR_ACCURACY = 2

class CakeEater:
    __slots__ = ('tummy_size', 'type')
    
    def __init__(self, type_, tummy_size):
        self.type = type_
        self.tummy_size = tummy_size
    
    def __repr__(self):
        return f'{self.__class__.__name__}({self.type!r}, {self.tummy_size:.{TUMMY_REPR_ACCURACY}f})'
    
    async def throw(self):
        # No one would ever expect this.
        raise StopIteration()

print(CakeEater('dream eater', 2.111))
```
"""
__all__ = ('HIGHLIGHT_TOKEN_TYPES', 'set_highlight_html_class', )

from scarletio.utils.highlight import HIGHLIGHT_TOKEN_TYPES, HighlightFormatterContext

PATCHOULI_FORMATTER_CONTEXT = HighlightFormatterContext()
PATCHOULI_FORMATTER_CONTEXT.set_highlight_html_all()


PYTHON_IDENTIFIERS = {'python', 'py', 'sage', 'python3', 'py3'}


def set_highlight_html_class(token_type_identifier, html_class):
    """
    Sets html class formatting for the given node.
    
    Parameters
    ----------
    token_type_identifier : `int`
        The node's identifier.
    html_class : `None`, `str`
        The html class to set.
    
    Raises
    ------
    TypeError
        - If `token_type_identifier` was not given as `int`.
        - If `html_class` was not given neither as `None` nor as `str`.
    ValueError
        If `token_type_identifier` was not given as any of the predefined values. Check ``TOKEN_TYPES`` for more
            details.
    """
    PATCHOULI_FORMATTER_CONTEXT.set_highlight_html_class(token_type_identifier, html_class)



__all__ = ('InteractionMetadataApplicationCommandAutocomplete',)

from scarletio import copy_docs

from .base import InteractionMetadataBase
from .fields import (
    parse_id, parse_name, parse_options, put_id_into, put_name_into, put_options_into, validate_id, validate_name,
    validate_options
)


class InteractionMetadataApplicationCommandAutocomplete(InteractionMetadataBase):
    """
    Interaction metadata used when the interaction was triggered by an application command's auto completion.
    
    Parameters
    ----------
    id : `int`
        The represented application command's identifier number.
    
    name : `str`
        The represented application command's name.
    
    options : `None`, `tuple` of ``InteractionOption``
        Application command option representations. Like sub-command or parameter.
    """
    __slots__ = ('id', 'name', 'options')
    
    @classmethod
    @copy_docs(InteractionMetadataBase._create_empty)
    def _create_empty(cls):
        self = object.__new__(cls)
        self.id = 0
        self.name = ''
        self.options = None
        return self
    
    
    @copy_docs(InteractionMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.id = self.id
        new.name = self.name
        
        options = self.options
        if (options is not None):
            options = tuple(option.copy() for option in options)
        new.options = options
        
        return new
    
    
    @copy_docs(InteractionMetadataBase._set_attributes_from_keyword_parameters)
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        # id
        try:
            application_command_id = keyword_parameters.pop('id')
        except KeyError:
            pass
        else:
            self.id = validate_id(application_command_id)
        
        # name
        try:
            application_command_name = keyword_parameters.pop('name')
        except KeyError:
            pass
        else:
            self.name = validate_name(application_command_name)
        
        # options
        try:
            options = keyword_parameters.pop('options')
        except KeyError:
            pass
        else:
            self.options = validate_options(options)
        
        
        InteractionMetadataBase._set_attributes_from_keyword_parameters(self, keyword_parameters)
    
    
    @classmethod
    @copy_docs(InteractionMetadataBase.from_data)
    def from_data(cls, data, interaction_event):
        self = object.__new__(cls)
        self.id = parse_id(data)
        self.name = parse_name(data)
        self.options = parse_options(data)
        return self
    
    
    @copy_docs(InteractionMetadataBase.to_data)
    def to_data(self, *, defaults = False, interaction_event = None):
        data = {}
        put_id_into(self.id, data, defaults)
        put_name_into(self.name, data, defaults)
        put_options_into(self.options, data, defaults)
        return data
    
    
    @copy_docs(InteractionMetadataBase._put_attribute_representations_into)
    def _put_attribute_representations_into(self, repr_parts):
        # id
        repr_parts.append(' id=')
        repr_parts.append(repr(self.id))
        
        # name
        repr_parts.append(', name=')
        repr_parts.append(repr(self.name))
        
        # options
        options = self.options
        if (options is not None):
            repr_parts.append(', options=')
            repr_parts.append(repr(options))
        
        return True
    
    
    @copy_docs(InteractionMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options)
            
            for option in options:
                hash_value ^= hash(option)
        
        return hash_value
    
    
    @copy_docs(InteractionMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # id
        if self.id != other.id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        return True



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, force_string_parser_factory, nullable_string_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, force_string_putter_factory, nullable_string_putter_factory,
    preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, force_string_validator_factory, nullable_string_validator_factory,
    preinstanced_validator_factory
)

from .preinstanced import OperationSystem

# launcher

parse_launcher = bool_parser_factory('is_launcher', False)
put_launcher_into = bool_optional_putter_factory('is_launcher', False)
validate_launcher = bool_validator_factory('launcher')

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', 0, 1024)

# os

parse_os = preinstanced_parser_factory('os', OperationSystem, OperationSystem.none)
put_os_into = preinstanced_putter_factory('os')
validate_os = preinstanced_validator_factory('os', OperationSystem)

# parameters

parse_parameters = nullable_string_parser_factory('parameters')
put_parameters_into = nullable_string_putter_factory('parameters')
validate_parameters = nullable_string_validator_factory('parameters', 0, 1024)



__all__ = ()

from ...field_parsers import (
    int_parser_factory, nullable_string_array_parser_factory, preinstanced_array_parser_factory
)
from ...field_putters import (
    int_putter_factory, nullable_string_array_optional_putter_factory, preinstanced_array_putter_factory
)
from ...field_validators import (
    int_conditional_validator_factory, nullable_string_array_validator_factory, preinstanced_array_validator_factory
)

from .constants import AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
from .preinstanced import AutoModerationKeywordPresetType

# excluded_keywords

parse_excluded_keywords = nullable_string_array_parser_factory('allow_list')
put_excluded_keywords_into = nullable_string_array_optional_putter_factory('allow_list')
validate_excluded_keywords = nullable_string_array_validator_factory('excluded_keywords')

# keyword_presets

parse_keyword_presets = preinstanced_array_parser_factory('presets', AutoModerationKeywordPresetType)
put_keyword_presets_into = preinstanced_array_putter_factory('presets')
validate_keyword_presets = preinstanced_array_validator_factory('keyword_presets', AutoModerationKeywordPresetType)

# keywords

parse_keywords = nullable_string_array_parser_factory('keyword_filter')
put_keywords_into = nullable_string_array_optional_putter_factory('keyword_filter')
validate_keywords = nullable_string_array_validator_factory('keywords')

# mention_limit

parse_mention_limit = int_parser_factory('mention_total_limit', AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX)
put_mention_limit_into = int_putter_factory('mention_total_limit')
validate_mention_limit = int_conditional_validator_factory(
    'mention_limit',
    AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX,
    (
        lambda mention_limit:
        mention_limit >= 0 and mention_limit <= AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
    ),
    f'>= {0} and <= {AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX},'
)

# regex_patterns

parse_regex_patterns = nullable_string_array_parser_factory('regex_patterns')
put_regex_patterns_into = nullable_string_array_optional_putter_factory('regex_patterns')
validate_regex_patterns = nullable_string_array_validator_factory('regex_patterns')



__all__ = ()

from .preinstanced import ButtonStyle, TextInputStyle


BUTTON_STYLE_DEFAULT = ButtonStyle.blue

TEXT_INPUT_STYLE_DEFAULT = TextInputStyle.short

LABEL_LENGTH_MAX = 80

PLACEHOLDER_LENGTH_MAX = 150

OPTIONS_LENGTH_MIN = 1
OPTIONS_LENGTH_MAX = 25

MAX_LENGTH_MIN = 0
MAX_LENGTH_MAX = 4000
MAX_LENGTH_DEFAULT = 0

MIN_LENGTH_MIN = 0
MIN_LENGTH_MAX = 4000
MIN_LENGTH_DEFAULT = 0

MAX_VALUES_MIN = 1
MAX_VALUES_MAX = 25
MAX_VALUES_DEFAULT = 1

MIN_VALUES_MIN = 0
MIN_VALUES_MAX = 15
MIN_VALUES_DEFAULT = 1

VALUE_LENGTH_MAX = 4000



__all__ = ('StringSelectOption',)

import reprlib

from scarletio import RichAttributeErrorBaseType

from ..shared_fields import parse_emoji, put_emoji_into, validate_emoji

from .fields import (
    parse_default, parse_description, parse_label, parse_value, put_default_into, put_description_into, put_label_into,
    put_value_into, validate_default, validate_description, validate_label, validate_value
)


class StringSelectOption(RichAttributeErrorBaseType):
    """
    An option of a string select component.
    
    Attributes
    ----------
    default : `bool`
        Whether this option is the default one.
    
    description : `None`, `str`
        Description of the option.
    
    emoji : `None`, ``Emoji``
        Emoji on the option if applicable.
    
    label : `str`
        Label of the option.
    
    value : `str`
        Identifier value of the option.
    """
    def __new__(cls, value, label = None, emoji = None, *, default = False, description = None):
        """
        Creates a new component option with the given parameters.
        
        Parameters
        ----------
        value : `str`
            The option's value.
        
        label : `str` = `None`, Optional
            Label of the component option.
            
            Defaults to the `value` parameter if not given.
        
        emoji : `None`, ``Emoji`` = `None`, Optional
            Emoji of the option if applicable.
        
        default : `bool` = `False`, Optional (Keyword only)
            Whether this the the default option. Defaults to `False`.
        
        description : `None`, `str` = `None`, Optional (Keyword only)
            Description of the component option.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        default = validate_default(default)
        description = validate_description(description)
        emoji = validate_emoji(emoji)
        label = validate_label(label)
        value = validate_value(value)
        
        if label is None:
            label = value
        
        self = object.__new__(cls)
        self.default = default
        self.description = description
        self.emoji = emoji
        self.label = label
        self.value = value
        return self


    @classmethod
    def from_data(cls, data):
        """
        Creates a new string select option from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            String select option data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        
        self.default = parse_default(data)
        self.description = parse_description(data)
        self.emoji = parse_emoji(data)
        self.label = parse_label(data)
        self.value = parse_value(data)
        
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the string select option to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
        """
        data = {}
        
        put_default_into(self.default, data, defaults)
        put_description_into(self.description, data, defaults)
        put_emoji_into(self.emoji, data, defaults)
        put_label_into(self.label, data, defaults)
        put_value_into(self.value, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns the string select option's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # System fields : value
        
        # value
        repr_parts.append(', value = ')
        repr_parts.append(reprlib.repr(self.value))
        
        # Text fields : emoji & label
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            repr_parts.append(', emoji = ')
            repr_parts.append(repr(emoji))
        
        # label
        label = self.label
        if (label is not None):
            repr_parts.append(', label = ')
            repr_parts.append(reprlib.repr(label))
        
        # Optional descriptive fields: description & default
        
        # description
        description = self.description
        if (description is not None):
            repr_parts.append(', description = ')
            repr_parts.append(reprlib.repr(description))
        
        # default
        if self.default:
            repr_parts.append(', default = True')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two string select options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # default
        if self.default != other.default:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # emoji
        if self.emoji is not other.emoji:
            return False
        
        # label
        if self.label != other.label:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the string select option's hash value."""
        hash_value = 0
        
        # default
        if self.default:
            hash_value ^= 1 << 8
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= emoji.id
        
        # label
        label = self.label
        if (label is not None):
            hash_value ^= hash(label)
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the string select option.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.default = self.default
        new.description = self.description
        new.emoji = self.emoji
        new.label = self.label
        new.value = self.value
        return new
    
    
    def copy_with(self, **keyword_parameters):
        """
        Copes the string select with modifying it's defined attributes.
        
        **keyword_parameters : Keyword parameters
            Keyword parameters defining which attribute should be overwritten.
        
        Other Parameters
        ----------------
        default : `bool`
            Whether this the the default option. Defaults to `False`.
        
        description : `None`, `str`, Optional (Keyword only)
            Description of the component option.
        
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            Emoji of the option if applicable.
        
        label : `str`, Optional (Keyword only)
            Label of the component option.
        
        value : `str`, Optional (Keyword only)
            The option's value.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        # default
        try:
            default = keyword_parameters.pop('default')
        except KeyError:
            default = self.default
        else:
            default = validate_default(default)
        
        # description
        try:
            description = keyword_parameters.pop('description')
        except KeyError:
            description = self.description
        else:
            description = validate_description(description)
        
        # emoji
        try:
            emoji = keyword_parameters.pop('emoji')
        except KeyError:
            emoji = self.emoji
        else:
           emoji = validate_emoji(emoji)
        
        
        # label
        try:
            label = keyword_parameters.pop('label')
        except KeyError:
            label = self.label
        else:
            label = validate_label(label)
        
        # value
        try:
            value = keyword_parameters.pop('value')
        except KeyError:
            value = self.value
        else:
            value = validate_value(value)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        if label is None:
            label = value
        
        new = object.__new__(type(self))
        new.default = default
        new.description = description
        new.emoji = emoji
        new.label = label
        new.value = value
        return new



__all__ = ('ApplicationType',)

from ...bases import Preinstance as P, PreinstancedBase


class ApplicationType(PreinstancedBase):
    """
    Represents an application type.
    
    Attributes
    ----------
    name : `str`
        The name of the application type.
    value : `int`
        The Discord side identifier value of the application type.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationType``) items
        Stores the created application type instances. This container is accessed when translating a Discord
        application type's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The application types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the application types.
    
    Every predefined application type can be accessed as class attribute as well:
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | game                      | game                      | 1     |
    +---------------------------+---------------------------+-------+
    | music                     | music                     | 2     |
    +---------------------------+---------------------------+-------+
    | ticketed_event            | ticketed event            | 3     |
    +---------------------------+---------------------------+-------+
    | guild_role_subscription   | guild role subscription   | 4     |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    game = P(1, 'game')
    music = P(2, 'music')
    ticketed_event = P(3, 'ticketed event')
    guild_role_subscription = P(4, 'guild role subscription')



__all__ = ('ChannelMetadataGuildStage',)

from scarletio import copy_docs

from ...permission import Permission
from ...permission.permission import (
    PERMISSION_MASK_CONNECT, PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE, PERMISSION_TEXT_DENY,
    PERMISSION_VOICE_DENY_CONNECTION
)

from .fields import parse_topic, put_topic_into, validate_topic

from .guild_main_base import ChannelMetadataGuildMainBase
from .guild_voice_base import ChannelMetadataGuildVoiceBase


class ChannelMetadataGuildStage(ChannelMetadataGuildVoiceBase):
    """
    Guild voice channel metadata.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    bitrate : `int`
        The bitrate (in bits) of the voice channel.
    region : ``VoiceRegion``
        The voice region of the channel.
    user_limit : `int`
        The maximal amount of users, who can join the voice channel, or `0` if unlimited.
    topic : `None`, `str`
        The channel's topic.
    
    Class Attributes
    ----------------
    order_group: `int` = `2`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('topic',)
    
    @copy_docs(ChannelMetadataGuildVoiceBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildVoiceBase.__hash__(self)
        
        # topic
        topic = self.topic
        if (topic is not None):
            hash_value ^= hash(topic)
        
        return hash_value
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildVoiceBase._is_equal_same_type(self, other):
            return False
        
        # topic
        if self.topic != other.topic:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildVoiceBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildStage, cls)._create_empty()
        
        self.topic = None
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildVoiceBase._update_attributes(self, data)
        
        # topic
        self.topic = parse_topic(data)
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildVoiceBase._difference_update_attributes(self, data)
        
        # topic
        topic = parse_topic(data)
        if self.topic != topic:
            old_attributes['topic'] = self.topic
            self.topic = topic
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._set_attributes_from_keyword_parameters)
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        ChannelMetadataGuildVoiceBase._set_attributes_from_keyword_parameters(self, keyword_parameters)
        
        # topic
        try:
            topic = keyword_parameters.pop('topic')
        except KeyError:
            pass
        else:
            self.topic = validate_topic(topic)
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildVoiceBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # topic
        put_topic_into(self.topic, data, defaults)
        
        return data
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # voice channels don't have text permissions
        result &= PERMISSION_TEXT_DENY
        
        if not result & PERMISSION_MASK_CONNECT:
            result &= PERMISSION_VOICE_DENY_CONNECTION
        
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        result = self._get_base_permissions_for_roles(channel_entity, roles)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # voice channels don't have text permissions
        result &= PERMISSION_TEXT_DENY
        
        if not result & PERMISSION_MASK_CONNECT:
            result &= PERMISSION_VOICE_DENY_CONNECTION
        
        return Permission(result)



__all__ = ()

from .scheduled_event import convert_privacy_level
from .shared import convert_nothing


STAGE_CONVERTERS = {
    'name': convert_nothing,
    'privacy_level': convert_privacy_level,
    'topic': convert_nothing,
}



__all__ = ()

from ....preconverters import preconvert_preinstanced_type

from ..helpers import get_permission_overwrite_key_value
from ..preinstanced import PermissionOverwriteTargetType


def parse_target_type(data):
    """
    Parses out ``PermissionOverwrite.target_type`` from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Permission overwrite data.
    
    Returns
    -------
    target_type : ``PermissionOverwriteTargetType``
    """
    return PermissionOverwriteTargetType.get(get_permission_overwrite_key_value(data))


def validate_target_type(target_type):
    """
    Validates the given `target_type` field of ``PermissionOverwrite``.
    
    Parameters
    ----------
    target_type : ``PermissionOverwriteTargetType``, `int`
        The permission overwrite's target type.
    
    Returns
    -------
    target_type : ``PermissionOverwriteTargetType``
    
    Raises
    ------
    TypeError
        - If `target_type` is not ``PermissionOverwriteTargetType``, `int`,
    """
    return preconvert_preinstanced_type(target_type, 'target_type', PermissionOverwriteTargetType)


def put_target_type_into(target_type, data, defaults):
    """
    Puts the `target` field of ``PermissionOverwrite`` to the given data.
    
    Parameters
    ----------
    target_type : ``PermissionOverwriteTargetType``
        The permission overwrite's target type.
    data : `dict` of (`str`, `Any`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    data['type'] = target_type.value
    
    return data



from .emoji import *
from .emoji_all_pattern import *
from .event_types import *
from .reaction_mapping import *
from .reaction_mapping_line import *
from .unicode_type import *
from .unicodes import *
from .utils import *

__all__ = (
    *emoji.__all__,
    *emoji_all_pattern.__all__,
    *event_types.__all__,
    *reaction_mapping.__all__,
    *reaction_mapping_line.__all__,
    *unicode_type.__all__,
    *unicodes.__all__,
    *utils.__all__,
)


from ...utils.module_deprecation import deprecated_import

deprecated_import(ReactionMapping, 'reaction_mapping')
deprecated_import(ReactionMappingLine, 'reaction_mapping_line')



__all__ = ()

from ...channel import Channel
from ...field_parsers import (
    default_entity_parser_factory, entity_id_parser_factory, nullable_string_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    entity_id_optional_putter_factory, entity_putter_factory, nullable_string_putter_factory,
    preinstanced_putter_factory
)
from ...field_validators import (
    default_entity_validator, entity_id_validator_factory, nullable_string_validator_factory,
    preinstanced_validator_factory
)
from ...message import Message
from ...user import ClientUserBase

from ..action import AutoModerationAction
from ..rule import AutoModerationRule, AutoModerationRuleTriggerType

# action

parse_action = default_entity_parser_factory('action', AutoModerationAction, AutoModerationAction())
put_action_into = entity_putter_factory('action', AutoModerationAction)
validate_action = default_entity_validator('action', AutoModerationAction, AutoModerationAction())

# alert_system_message_id

parse_alert_system_message_id = entity_id_parser_factory('alert_system_message_id')
put_alert_system_message_id_into = entity_id_optional_putter_factory('alert_system_message_id')
validate_alert_system_message_id = entity_id_validator_factory('alert_system_message_id', Message)

# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id_into = entity_id_optional_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', Channel)

# content

parse_content = nullable_string_parser_factory('content')
put_content_into = nullable_string_putter_factory('content')
validate_content = nullable_string_validator_factory('content', 0, 10000)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', include = 'Guild')

# matched_content

parse_matched_content = nullable_string_parser_factory('matched_content')
put_matched_content_into = nullable_string_putter_factory('matched_content')
validate_matched_content = nullable_string_validator_factory('matched_content', 0, 10000)

# matched_keyword

parse_matched_keyword = nullable_string_parser_factory('matched_keyword')
put_matched_keyword_into = nullable_string_putter_factory('matched_keyword')
validate_matched_keyword = nullable_string_validator_factory('matched_keyword', 0, 10000)

# rule_id

parse_rule_id = entity_id_parser_factory('rule_id')
put_rule_id_into = entity_id_optional_putter_factory('rule_id')
validate_rule_id = entity_id_validator_factory('rule_id', AutoModerationRule)

# rule_trigger_type

parse_rule_trigger_type = preinstanced_parser_factory(
    'rule_trigger_type', AutoModerationRuleTriggerType, AutoModerationRuleTriggerType.none
)
put_rule_trigger_type_into = preinstanced_putter_factory('rule_trigger_type')
validate_rule_trigger_type = preinstanced_validator_factory('rule_trigger_type', AutoModerationRuleTriggerType)

# user_id

parse_user_id = entity_id_parser_factory('user_id')
put_user_id_into = entity_id_optional_putter_factory('user_id')
validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)



__all__ = ()

import warnings

from scarletio import RichAttributeErrorBaseType

from ..sticker import StickerFormat


STICKER_FORMAT_STATIC = StickerFormat.png
STICKER_FORMAT_ANIMATED = StickerFormat.apng
STICKER_FORMAT_LOTTIE = StickerFormat.lottie


class StickerCounts(RichAttributeErrorBaseType):
    """
    Represents a guild's counted stickers by their type.
    
    Attributes
    ----------
    animated : `int`
        Animated sticker count.
    lottie : `int`
        Lottie sticker count.
    static : `int`
        Static sticker count.
    
    Notes
    -----
    Use ``Guild.sticker_limit`` to get it's sticker limit.
    """
    __slots__ = ('animated', 'lottie', 'static')
    
    def __new__(
        cls,
        *,
        animated = 0,
        lottie = 0,
        static = 0,
    ):
        """
        Creates a new sticker counts with the given fields.
        
        Parameters
        ----------
        animated : `int` = `0`, Optional (Keyword only)
            Animated sticker count.
        lottie : `int` = `0`, Optional (Keyword only)
            Lottie sticker count.
        static : `int` = `0`, Optional (Keyword only)
            Static sticker count.
        """
        self = object.__new__(cls)
        self.animated = animated
        self.lottie = lottie
        self.static = static
        return self
    
    
    @classmethod
    def from_stickers(cls, iterator):
        """
        Creates a new sticker counts instance from the iterable.
        
        Parameters
        ----------
        iterator : `iterable` of ``Sticker``
            Sticker iterator.
        
        Returns
        -------
        self : `instance<cls>`
        """
        animated = 0
        lottie = 0
        static = 0
        
        for sticker in iterator:
            sticker_format = sticker.format
            if sticker_format is STICKER_FORMAT_ANIMATED:
                animated += 1
                continue
            
            if sticker_format is STICKER_FORMAT_LOTTIE:
                lottie += 1
                continue
            
            if sticker_format is STICKER_FORMAT_STATIC:
                static += 1
                continue
        
        
        self = object.__new__(cls)
        self.animated = animated
        self.lottie = lottie
        self.static = static
        return self
    
    
    @property
    def normal_total(self):
        """
        Returns the total amount of normal stickers.
        
        Returns
        -------
        normal_total : `int`
        """
        return self.animated + self.static
    
    @property
    def total(self):
        """
        Returns the total count of stickers.
        
        Returns
        -------
        total : `int`
        """
        return self.lottie + self.static + self.animated
    
    
    def _iter_field_values(self):
        """
        Iterates over the fields of the sticker counts.
        
        This method is an iterable generator.
        
        Yields
        ------
        field_value : `int`
        """
        yield self.animated
        yield self.lottie
        yield self.static
    
    
    def __bool__(self):
        """Returns whether the sticker counts counted any stickers"""
        for field_value in self._iter_field_values():
            if field_value:
                return True
        
        return False
    
    
    def __repr__(self):
        """Returns the sticker counts' representation"""
        repr_parts = [self.__class__.__name__, '(']
        field_added = False
        
        for field_name, field_value in zip(self.__slots__, self._iter_field_values()):
            if field_value:
                if field_added:
                    repr_parts.append(', ')
                else:
                    field_added = True
                
                repr_parts.append(field_name)
                repr_parts.append(' = ')
                repr_parts.append(repr(field_value))
        
        repr_parts.append(')')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two sticker counts equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        for self_field, other_field in zip(self._iter_field_values(), other._iter_field_values()):
            if self_field != other_field:
                return False
        
        return True
    
    
    def __hash__(self):
        """Returns the sticker counts' hash value."""
        hash_value = 0
        
        shift = 0
        
        for field_value in self._iter_field_values():
            hash_value ^= field_value << shift
            shift += 4
        
        return hash_value
    
    
    def __iter__(self):
        """Deprecated and will be removed in 2023 April."""
        warnings.warn(
            (
                f'Unpacking `Guild.sticker_counts` is deprecated and will be removed in 2023 April.\n'
                f'Please do the following instead:\n'
                f'sticker_counts = guild.sticker_counts\n'
                f'animated = sticker_counts.animated\n'
                f'lottie = sticker_counts.lottie\n'
                f'static = sticker_counts.static'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        yield self.static
        yield self.animated
        yield self.lottie



__all__ = ('RoleManagerMetadataBase',)

from scarletio import RichAttributeErrorBaseType

from ...bases import PlaceHolder


class RoleManagerMetadataBase(RichAttributeErrorBaseType):
    """
    Base class for role managers.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new role manager.
        """
        return object.__new__(cls)
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new role manager metadata from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Role manager metadata.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the role manager metadata to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        return {}
    
    
    def __repr__(self):
        """Returns the the role manager metadata's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    def __hash__(self):
        """Returns the role manager metadata's hash."""
        return 0
    
    
    def __eq__(self, other):
        """Returns whether the two role manager metadatas are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return True
    
    
    def copy(self):
        """
        Copies the role manager metadata.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return object.__new__(type(self))
    
    
    def copy_with(self):
        """
        Copies the role manager metadata with the given fields.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return self.copy()
    
    
    @property
    def manager_id(self):
        """
        Returns the entity's identifier to which the role is managed by.
        
        Returns
        -------
        manager_id : `int`
        """
        return 0
    
    
    @property
    def manager(self):
        """
        Returns the entity's identifier to which the role is managed by.
        
        Returns
        -------
        manager : `None`, ``ClientUserBase``, ``Integration``
        """
        return None
    
    
    # place-holders
    
    bot_id = PlaceHolder(
        0,
        """
        The bot's identifier to which the role belongs to.
        
        Returns
        -------
        bot_id : `int`
        """
    )
    
    
    integration_id = PlaceHolder(
        0,
        """
        The integration's identifier to which the role belongs to.
        
        Returns
        -------
        integration_id : `int`
        """
    )
    
    
    purchasable = PlaceHolder(
        False,
        """
        Whether this role is available for purchase.
        
        Returns
        -------
        purchasable : `bool`
        """
    )
    
    
    subscription_listing_id = PlaceHolder(
        0,
        """
        The subscription listing's and sku's identifier to which the role belongs to.
        
        Returns
        -------
        subscription_listing_id : `int`
        """
    )



from .import_overwrite import *
from .snapshot import *
from .utils import *

from .client_extension import *
from .constants import *
from .exceptions import *
from .plugin import *
from .plugin_loader import *
from .plugin_root import *
from .helpers import *


__all__ = (
    *import_overwrite.__all__,
    *snapshot.__all__,
    *utils.__all__,
    
    *client_extension.__all__,
    *constants.__all__,
    *exceptions.__all__,
    *plugin.__all__,
    *plugin_loader.__all__,
    *plugin_root.__all__,
    *helpers.__all__,
)

from .. import register_library_extension
register_library_extension('HuyaneMatsu.plugin_loader')
del register_library_extension



__all__ = ('ReactionMapping', )

from scarletio import set_docs
from scarletio.utils.compact import NEEDS_DUMMY_INIT

from .emoji import Emoji
from .reaction_mapping_line import ReactionMappingLine
from .utils import create_partial_emoji_data, create_partial_emoji_from_data


def _validate_reaction_mapping_initialize_with(reaction_mapping_type, initialize_with):
    """
    Validates the `initialize_with` parameter of ``ReactionMapping``.
    
    Parameters
    ----------
    reaction_mapping_type : `type`
        The type of the reaction mapping.
    initialize_with : `None`, `iterable` of `tuple` (``Emoji``, `iterable` of (``ClientUserBase``, `None`)), `dict` of \
            (``Emoji``, `iterable` of (``ClientUserBase``, `None`) items, `instance<reaction_mapping_type>`
        The value to initialise the reaction mapping.
    
    Returns
    -------
    built_initialize_with : `None`, `list` of `tuple` (``Emoji``, ``ReactionMappingLine``) items
        The validated extend with value.
    
    Raises
    ------
    TypeError
        - If `initialize_with`'s type is unacceptable.
        - If an item (or element) of `initialize_with` has incorrect type, length or structure.
    """
    if initialize_with is None:
        built_initialize_with = None
    
    elif type(initialize_with) is reaction_mapping_type:
        built_initialize_with = _validate_reaction_mapping_initialize_with_same(initialize_with)
    
    elif isinstance(initialize_with, dict):
        built_initialize_with = _validate_reaction_mapping_initialize_with_dict(initialize_with)
    
    elif (getattr(initialize_with, '__iter__', None) is not None):
        built_initialize_with = _validate_reaction_mapping_initialize_with_iterable(initialize_with)
    
    else:
        raise TypeError(
            f'`initialize_with` can be `None`, `{reaction_mapping_type.__name__}`, `iterable` or `dict`, got '
            f'{initialize_with.__class__.__name__}; {initialize_with!r}'
        )
    
    return built_initialize_with


def _validate_reaction_mapping_initialize_with_same(initialize_with):
    """
    Validates the `initialize_with` parameter of ``ReactionMapping`` where it is the same type.
    
    Parameters
    ----------
    initialize_with : ``ReactionMapping``
        The value to initialize_with self with.
    
    Returns
    -------
    built_initialize_with : `None`, `list` of `tuple` (``Emoji``, ``ReactionMappingLine``) items
        The validated extend with value.
    """
    built_initialize_with = None
    
    for emoji, line in initialize_with.items():
        if built_initialize_with is None:
            built_initialize_with = []
        
        built_initialize_with.append((emoji, line.copy()))
    
    return built_initialize_with


def _validate_reaction_mapping_initialize_with_dict(initialize_with):
    """
    Validates the `initialize_with` parameter of ``ReactionMapping`` where it is a dictionary.
    
    Parameters
    ----------
    initialize_with : `dict` of (``Emoji``, `iterable` of (``ClientUserBase``, `None`) items
        The value to initialize_with self with.
    
    Returns
    -------
    built_initialize_with : `None`, `list` of `tuple` (``Emoji``, ``ReactionMappingLine``) items
        The validated extend with value.
    
    Raises
    ------
    TypeError
        - If `initialize_with` has an item of an unacceptable structure.
    """
    built_initialize_with = None
    for item in initialize_with.items():
        built_initialize_with = _validate_reaction_mapping_initialize_with_item(built_initialize_with, item)
    
    return built_initialize_with


def _validate_reaction_mapping_initialize_with_iterable(initialize_with):
    """
    Validates the `initialize_with` parameter of ``ReactionMapping`` where it is any iterable.
    
    Parameters
    ----------
    initialize_with : `iterable` of `tuple` (``Emoji``, `iterable` of (``ClientUserBase``, `None`))
        The value to initialize_with self with.
    
    Returns
    -------
    built_initialize_with : `None`, `list` of `tuple` (``Emoji``, ``ReactionMappingLine``) items
        The validated extend with value.
    
    Raises
    ------
    TypeError
        - If `initialize_with` contains a non `tuple` element.
        - If `initialize_with` contains an element which length is not `2`.
        - If `initialize_with` has an element of an unacceptable structure.
    """
    built_initialize_with = None
    
    for item in initialize_with:
        if not isinstance(item, tuple):
            raise TypeError(
                f'`initialize_with` items can be `tuple` instances, got '
                f'{item.__class__.__name__}; {item!r}; initialize_with{initialize_with!r}'
            )
        
        item_length = len(item)
        if len(item) != 2:
            raise TypeError(
                f'`initialize_with` items can be `tuple` with length of `2`, got '
                f'item_length={item_length!r}; item={item!r}; initialize_with{initialize_with!r}'
            )
        
        built_initialize_with = _validate_reaction_mapping_initialize_with_item(built_initialize_with, item)
    
    return built_initialize_with


def _validate_reaction_mapping_initialize_with_item(built_initialize_with, item):
    """
    Validates an item of the `initialize_with` parameter of ``ReactionMapping``.
    
    Parameters
    ----------
    built_initialize_with : `None`, `list` of `tuple` (``Emoji``, ``ReactionMappingLine``) items
        The validated extend with value.
    item : `tuple` (``Emoji``, `iterable` of (``ClientUserBase``, `None`))
        Reaction mapping item to validate.
    
    Returns
    -------
    built_initialize_with : `None`, `list` of `tuple` (``Emoji``, ``ReactionMappingLine``) items
        The validated extend with value.
    
    Raises
    ------
    TypeError
        - If `item[0]` is not ``Emoji``.
        - If `item[1]` isn ot accepted by ``ReactionMappingLine``.
    """
    emoji, line = item
    
    if not isinstance(emoji, Emoji):
        raise TypeError(
            f'`item[0]` can be `{Emoji.__name__}` instance, got '
            f'{emoji.__class__.__name__}; {emoji!r}; item={item!r}.'
        )
    
    line = ReactionMappingLine(line)
    
    if line:
        if built_initialize_with is None:
            built_initialize_with = []
        
        built_initialize_with.append((emoji, line))
    
    return built_initialize_with


class ReactionMapping(dict):
    """
    A `dict` subclass, which contains the reactions on a ``Message`` with (``Emoji``, ``ReactionMappingLine``)
    items.
    
    Attributes
    ----------
    fully_loaded : `bool`
        Whether the reaction mapping line is fully loaded.
    """
    __slots__ = ('fully_loaded',)
    
    if NEEDS_DUMMY_INIT:
        def __init__(self, *args, **kwargs):
            pass
    else:
        __init__ = object.__init__
    
    
    def __new__(cls, initialize_with=None):
        """
        Creates a new reaction mapping instance.
        
        Parameters
        ----------
        initialize_with : `None`, `iterable` of `tuple` (``Emoji``, `iterable` of (``ClientUserBase``, `None`)), `dict` of \
                (``Emoji``, `iterable` of (``ClientUserBase``, `None`) items, `instance<cls>` = `None`, Optional
            The value to initialize_with self with.
        
        Raises
        ------
        TypeError
            - If `initialize_with`'s is unacceptable.
        """
        built_initialize_with = _validate_reaction_mapping_initialize_with(cls, initialize_with)
        
        self = dict.__new__(cls)
        self.fully_loaded = True
        
        if (built_initialize_with is not None):
            dict.update(self, built_initialize_with)
            self._full_check()
        
        return self
    
    
    def __bool__(self):
        """Returns whether self has any any reactions"""
        if dict.__len__(self):
            return True
        
        return False
    
    
    def __repr__(self):
        """Returns the reaction mapping's representation."""
        repr_parts = [self.__class__.__name__, '(']
        
        if dict.__len__(self):
            repr_parts.append('{')
            for emoji, line in dict.items(self):
                repr_parts.append(repr(emoji))
                repr_parts.append(': ')
                repr_parts.append(repr(line))
                repr_parts.append(', ')
            
            repr_parts[-1] = '}'
        
        repr_parts.append(')')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether self equals to other."""
        if type(other) is type(self):
            return dict.__eq__(self, other)
        
        if isinstance(other, dict):
            return self._is_equal_dict(other)
        
        return NotImplemented
    
    
    def _is_equal_dict(self, other):
        """
        Returns whether self equals to the given dictionary.
        
        Parameters
        ----------
        other : `dict` of (``Emoji``, `iterable` of (``ClientUserBase``, `None`)) items
            The other instance to compare self to.
        
        Returns
        -------
        is_equal : `bool`, `NotImplemented`
        """
        for emoji in other.keys():
            if not isinstance(emoji, Emoji):
                return NotImplemented
        
        is_equal = True
        
        for emoji in {*dict.keys(self), *other.keys()}:
            try:
                users = other[emoji]
            except KeyError:
                is_equal = False
                continue
            
            try:
                line = self[emoji]
            except KeyError:
                line = ReactionMappingLine._create_empty(0)
                is_equal = False
            
            line_equals = type(line).__eq__(line, users)
            if line_equals is NotImplemented:
                return NotImplemented
            
            if not line_equals:
                is_equal = False
            continue
        
        return is_equal
    
    
    @classmethod
    def from_data(cls, data):
        """
        Fills the reaction mapping with the given data.
        
        Parameters
        ----------
        data : `None`, `dict` of (`str`, `Any`) items
            Reactions data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = dict.__new__(cls)
        if (data is None) or (not data):
            self.fully_loaded = True
            return self
        
        self.fully_loaded = False
        for line_data in data:
            emoji = create_partial_emoji_from_data(line_data['emoji'])
            line = ReactionMappingLine._create_empty(line_data.get('count', 1))
            self[emoji] = line
        
        return self
    
    
    emoji_count = set_docs(
        property(dict.__len__),
        """
        The amount of different emojis, which were added on the reaction mapping's respective ``Message``.
        
        Returns
        -------
        emoji_count : `int`
        """
    )
    
    @property
    def total_count(self):
        """
        The total amount reactions given on the reaction mapping's respective message.
        
        Returns
        -------
        total_count : `int`
        """
        total_reactions = 0
        
        for users in dict.values(self):
            total_reactions += len(users)
        
        return total_reactions
    
    
    def clear(self):
        """
        Clears the reaction mapping with clearing it's lines.
        """
        for value in self.values():
            value.clear()
        
        if self.fully_loaded:
            self._full_check()
    
    
    def add(self, emoji, user):
        """
        Adds a user to the reactors.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The reacted emoji.
        user : ``ClientUserBase``
            The reactor user.
        """
        try:
            line = self[emoji]
        except KeyError:
            line = ReactionMappingLine._create_empty(0)
            self[emoji] = line
        
        line.add(user)
    
    
    def remove(self, emoji, user):
        """
        Removes a user to the reactors.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The removed reacted emoji.
        user : ``ClientUserBase``
            The removed reactor user.
        
        Returns
        -------
        success : `bool`
        """
        try:
            line = self[emoji]
        except KeyError:
            return False
        
        success = line.remove(user)
        if success:
            if not line:
                del self[emoji]
        
        return success
    
    
    def remove_emoji(self, emoji):
        """
        Removes all the users who reacted with the given ``Emoji`` and then returns the stored line.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The emoji to remove.
        
        Returns
        -------
        line : `None`, ``ReactionMappingLine``
        """
        line = self.pop(emoji, None)
        if (line is not None) and line.unknown:
            self._full_check()
        
        return line
    
    
    # this function is called if an emoji loses all it's unknown reactors
    def _full_check(self):
        """
        Checks whether the reaction mapping is fully loaded, by checking it's values' `.unknown` and sets the current
        state to `.fully_loaded`.
        """
        for line in self.values():
            if line.unknown:
                self.fully_loaded = False
                return
        
        self.fully_loaded = True
    
    
    # we call this when we get SOME reactors of an emoji
    def _update_some_users(self, emoji, users):
        """
        Called when some reactors of an emoji are updated.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The emoji, which users' are updated.
        users : `list` of ``ClientUserBase``
            The added reactors.
        """
        try:
            line = self[emoji]
        except KeyError:
            line = ReactionMappingLine._create_empty(0)
            self[emoji] = line
            
        line.update(users)
        self._full_check()
    
    
    def _update_all_users(self, emoji, users):
        """
        Called when all the reactors of an emoji are updated of the reaction mapping.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The emoji, which users' are updated.
        users : `list` of ``ClientUserBase``
            The added reactors.
        """
        self[emoji] = ReactionMappingLine._create_full(users)
        self._full_check()
    
    
    def to_data(self):
        """
        Tries to convert the reactions back to a json serializable list.
        
        Returns
        -------
        data : `list` of `dict` of (`str`, `Any`)
        """
        data = []
        
        for emoji, users in dict.items(self):
            data.append({
                'count': len(users),
                'me': False, # Me is always False
                'emoji': create_partial_emoji_data(emoji),
            })
        
        return data
